/*
 * CheckRedFlag.h
 *
 * Description:
 * 	Routines for red flag detection.
 * 	Must call SetupQuadrants() to initialize data.
 *
 * Authors:
 * 	Tim Li
 */

#include "xparameters.h"
#include "xutil.h"
#include "xbasic_types.h"

/*
 * OLD: Absolute compare version of thresholding
 *
#define MIN_RED_THRESHOLD	(200)	//on red channel only
#define MAX_GREEN_THRESHOLD	(100)	
#define MAX_BLUE_THRESHOLD	(100)
*/

// (red-green) AND (red-blue) needs to be >= DELTA_RED_THRESHOLD
#define DELTA_RED_THRESHOLD	(50)

//quadrant parameters
#define QUAD_ACROSS			(8)
#define QUAD_HIGH			(6)
#define QUADRANT_COUNT		(48)	//MUST BE QUAD_ACROSS*QUAD_HIGH

//quadrant locations
	#define LEFT	(0)	//constant "names" to use
	#define TOP		(1)	//  for referring to sides
	#define RIGHT	(2)	//  in the array structure below
	#define BOTTOM	(3)
	#define SIDES	(4)
long CHECK_QUAD_RECTS[QUADRANT_COUNT][SIDES];

//pixel checking parameters
#define MIN_SIZE			(5)		//need at least this many hits with below spacing in both X/Y directions
#define CHECK_SPACING		(3)		//number of pixels to skip in X/Y scans, to speed things up!
									// USE AN ODD NUMBER to utilize both odd and even fields of video input
const long MIN_QUAD_SCORE	= MIN_SIZE*MIN_SIZE;
	//minimum of 25 (5*5) pixel hits to consider having sufficient red

//color component value extraction macros
#define RED_COLOR_MASK			(0x00FF0000)
#define RED_COLOR_DOWNSHIFT		(16)
#define GREEN_COLOR_MASK		(0x0000FF00)
#define GREEN_COLOR_DOWNSHIFT	(8)
#define BLUE_COLOR_MASK			(0x000000FF)
#define BLUE_COLOR_DOWNSHIFT	(0)
#define getRED(col)		((col & RED_COLOR_MASK)   >> RED_COLOR_DOWNSHIFT)
#define getGREEN(col)	((col & GREEN_COLOR_MASK) >> GREEN_COLOR_DOWNSHIFT)
#define getBLUE(col)	((col & BLUE_COLOR_MASK)  >> BLUE_COLOR_DOWNSHIFT)


//computes co-ordinates for sides of "quadrant" rectangles
void SetupQuadrants()
{
	const int W = WIDTH  / QUAD_ACROSS;
	const int H = HEIGHT / QUAD_HIGH;
	int rect_num = 0;
	int x, y;
	int scrx=0, scry=0;

	for (y=0; y<QUAD_HIGH; y++)
	{
		scrx = 0;
		for (x=0; x<QUAD_ACROSS; x++)
		{
			CHECK_QUAD_RECTS[rect_num][LEFT] = scrx;
			CHECK_QUAD_RECTS[rect_num][TOP] = scry;
			CHECK_QUAD_RECTS[rect_num][RIGHT] = scrx + W;
			CHECK_QUAD_RECTS[rect_num][BOTTOM] = scry + H;
			
			rect_num++;
			scrx += W;
		}
		scry += H;
	}
}

/*
 * OLD COMMENT BELOW.
 * NOW returns from 1-48 for the hit quadrant, or 0 for none found
 *  with 1-8 being the top line from left to right, 9-16 the second line, etc.
*/
//this function tries to identify the 9-sector "quadrant" a colored flag is located in
//  RETURNS: 0 for not found, 1-9 for the "quadrant" detected,
//										in the order	1 2 3
//													4 5 6
//													7 8 9
int LocateFlagSector()
{
	long *screen;
	long quadrant_score[QUADRANT_COUNT];	//record scores per quadrant here.
	int quadrant;
	int x, y;
	int color, red, green, blue;
	int best_quad;
	int best_score;

	//check each quadrant, adding one to the score for each pixel that meets the threshold
	for (quadrant=0; quadrant < QUADRANT_COUNT; quadrant++)
	{
		quadrant_score[quadrant] = 0;	//reset score
		
		for (y=CHECK_QUAD_RECTS[quadrant][TOP]; y<=CHECK_QUAD_RECTS[quadrant][BOTTOM]; y+=CHECK_SPACING)
		{
			screen = (long*)XPAR_ZBT_512KX32_MEM0_BASEADDR + y*640;
			for (x=CHECK_QUAD_RECTS[quadrant][LEFT]; x<=CHECK_QUAD_RECTS[quadrant][RIGHT]; x+=CHECK_SPACING)
			{
				color = screen[x];		//screen = start of line address
				red = getRED(color);
				green = getGREEN(color);
				blue = getBLUE(color);
				
				//OLD THRESHOLD COMPARISON
				//if (red >= MIN_RED_THRESHOLD && green <= MAX_GREEN_THRESHOLD && blue <= MAX_BLUE_THRESHOLD)
				//	quadrant_score[quadrant]++;
				
				//current comparison, if the red value is higher than both the green and blue values above a threshold
				if ((red-green) >= DELTA_RED_THRESHOLD && (red-blue) >= DELTA_RED_THRESHOLD)
					quadrant_score[quadrant]++;
			}
		}
	}

	//compare scores and determine most likely quadrant
	best_quad = 0;	//none found
	best_score = 0;
	for (quadrant=QUADRANT_COUNT-1; quadrant >= 0; quadrant--)
	{
		if (quadrant_score[quadrant] > MIN_QUAD_SCORE && quadrant_score[quadrant] > best_score)
		{	//ensure it is a greater score, and above the minimum needed
			best_score = quadrant_score[quadrant];
			best_quad = quadrant+1;	//return value is ONE-indexed
		}
	}

	return best_quad;
}


#define BOX_WIDTH 5
//draws a bounding box to show the user the quadrant we are detecting
void DrawBoxForQuadrant(int quadrant)
{
	rect(CHECK_QUAD_RECTS[quadrant-1][LEFT],  CHECK_QUAD_RECTS[quadrant-1][TOP],
		 CHECK_QUAD_RECTS[quadrant-1][RIGHT], CHECK_QUAD_RECTS[quadrant-1][TOP]+BOX_WIDTH,
		 makecol(0,0,255));
	rect(CHECK_QUAD_RECTS[quadrant-1][LEFT],  CHECK_QUAD_RECTS[quadrant-1][TOP],
		 CHECK_QUAD_RECTS[quadrant-1][LEFT]+BOX_WIDTH, CHECK_QUAD_RECTS[quadrant-1][BOTTOM],
		 makecol(0,0,255));
	rect(CHECK_QUAD_RECTS[quadrant-1][LEFT],  CHECK_QUAD_RECTS[quadrant-1][BOTTOM]-BOX_WIDTH,
		 CHECK_QUAD_RECTS[quadrant-1][RIGHT], CHECK_QUAD_RECTS[quadrant-1][BOTTOM],
		 makecol(0,0,255));
	rect(CHECK_QUAD_RECTS[quadrant-1][RIGHT]-BOX_WIDTH,  CHECK_QUAD_RECTS[quadrant-1][TOP],
		 CHECK_QUAD_RECTS[quadrant-1][RIGHT], CHECK_QUAD_RECTS[quadrant-1][BOTTOM],
		 makecol(0,0,255));
}
