/*
 * Game.h
 * 
 * Description:
 * 	Routines for running the game.
 * 	Call GameProcFrame() every game frame.
 *
 * Authors:
 * 	Leo Hwang
 */

//set the top-right corner quadrant as the "start game" location
#define GAME_INIT_SQUARE	7

//target parameters
#define MAX_TARGETS		32
#define TARGET_MIN_SPEED	2
#define TARGET_MAX_SPEED	11
#define TARGET_RADIUS	8
#define TARGET_COLOUR	0x0000ffff
int targets[32][3];		//x-coord, y-coord and speed for each target
int num_targets;

//lives parameters
int num_lives;
int score;
#define MAX_LIVES	5
#define LIFE_BAR_X			500
#define LIFE_BAR_Y			20
#define LIFE_BAR_SPACING	20
#define LIFE_BAR_COLOUR		0x00cccc00


//add a new target to the game
void add_target (int x, int y, int speed) {
	targets[num_targets][0] = x;
	targets[num_targets][1] = y;
	targets[num_targets][2] = speed;
	num_targets++;
}

//check if a target is within the user's flag quadrant
//  NOTE: CHECK_QUAD_RECTS is found in CheckRedFlag.h
int collide_target (int target, int quadrant) {
	if (targets[target][0] >= CHECK_QUAD_RECTS[quadrant][LEFT] &&
		targets[target][0] <= CHECK_QUAD_RECTS[quadrant][RIGHT] &&
		targets[target][1] >= CHECK_QUAD_RECTS[quadrant][TOP] &&
		targets[target][1] <= CHECK_QUAD_RECTS[quadrant][BOTTOM]) {
		return 1;
	}
	return 0;
}

//remove a target from the list
void remove_target (int target) {
	int i;
	num_targets -= 1;

	//For removing the last target, we are done.
	if (target == num_targets) {
		return;
	}

	//For removing a target in the middle, shift the rest down.
	for (i = target; i < num_targets; i++) {
		targets[i][0] = targets[i + 1][0];
		targets[i][1] = targets[i + 1][1];
		targets[i][2] = targets[i + 1][2];
	}
}

//draw the lives (as yellow squares) the user still has REMAINING in the top-right
void draw_lives () {
	int i, x;

	for (i = 0; i < num_lives - 1; i++) {
		x = LIFE_BAR_X + i * LIFE_BAR_SPACING * 2;
		rect(x, LIFE_BAR_Y, x + LIFE_BAR_SPACING - 1, LIFE_BAR_Y + LIFE_BAR_SPACING - 1, LIFE_BAR_COLOUR);
	}
}

//draw the targets on the screen
void draw_targets () {
	int i;

	for (i = 0; i < num_targets; i++) {
		rect(targets[i][0] - TARGET_RADIUS + 1,
			targets[i][1] - TARGET_RADIUS + 1,
			targets[i][0] + TARGET_RADIUS - 1,
			targets[i][1] + TARGET_RADIUS - 1,
			TARGET_COLOUR);
	}
}

//draw the user's score (number of targets cleared) in the top-left
#define DIGIT_DOT_SIZE  (5)
#define SCORE_COLOR 	makecol(255, 128, 0)
void draw_score()
{
	int digits[3];	//assume score <= 999
	
	digits[0] = score / 100;	//hundreds
	digits[1] = (score % 100) / 10;	//tens
	digits[2] = (score % 10);		//ones

	drawNumber(10, 10, digits[0], DIGIT_DOT_SIZE, SCORE_COLOR);
	drawNumber(30, 10, digits[1], DIGIT_DOT_SIZE, SCORE_COLOR);
	drawNumber(50, 10, digits[2], DIGIT_DOT_SIZE, SCORE_COLOR);
}

//run a frame of the game!
void GameProcFrame(int quadFoundRedFlag) {
	int target_x, target_y;
	int random_x, random_speed;
	int i;

	//if game not started, draw the "start game" indicator box on the top-right corner
	if (num_lives <= 0) {
		rect(CHECK_QUAD_RECTS[GAME_INIT_SQUARE][LEFT],  CHECK_QUAD_RECTS[GAME_INIT_SQUARE][TOP],
			 CHECK_QUAD_RECTS[GAME_INIT_SQUARE][RIGHT], CHECK_QUAD_RECTS[GAME_INIT_SQUARE][TOP]+BOX_WIDTH,
			 makecol(0,0,255));
		rect(CHECK_QUAD_RECTS[GAME_INIT_SQUARE][LEFT],  CHECK_QUAD_RECTS[GAME_INIT_SQUARE][TOP],
			 CHECK_QUAD_RECTS[GAME_INIT_SQUARE][LEFT]+BOX_WIDTH, CHECK_QUAD_RECTS[GAME_INIT_SQUARE][BOTTOM],
			 makecol(0,0,255));
		rect(CHECK_QUAD_RECTS[GAME_INIT_SQUARE][LEFT],  CHECK_QUAD_RECTS[GAME_INIT_SQUARE][BOTTOM]-BOX_WIDTH,
			 CHECK_QUAD_RECTS[GAME_INIT_SQUARE][RIGHT], CHECK_QUAD_RECTS[GAME_INIT_SQUARE][BOTTOM],
			 makecol(0,0,255));
		rect(CHECK_QUAD_RECTS[GAME_INIT_SQUARE][RIGHT]-BOX_WIDTH,  CHECK_QUAD_RECTS[GAME_INIT_SQUARE][TOP],
			 CHECK_QUAD_RECTS[GAME_INIT_SQUARE][RIGHT], CHECK_QUAD_RECTS[GAME_INIT_SQUARE][BOTTOM],
			 makecol(0,0,255));

		//check and see if the user hit that corner -- if so, start the game
		if (quadFoundRedFlag == GAME_INIT_SQUARE + 1) {
			num_lives = MAX_LIVES;
			num_targets = 0;
			score = 0;
		}
	}
	else {	//game is running
		//Making new targets
		if (num_targets < MAX_TARGETS) {
			if (rand() % 10 == 0) {			//Spawn a new target?
				random_x = (rand() % (WIDTH-TARGET_RADIUS)) + TARGET_RADIUS/2;	//Random x coordinate
				random_speed = rand() % (TARGET_MAX_SPEED - TARGET_MIN_SPEED + 1) + TARGET_MIN_SPEED;
				add_target(random_x, 0, random_speed);		//Add new target
			}
		}

		//Collision detection
		if (quadFoundRedFlag != 0) {
			for (i = num_targets - 1; i >= 0; i--) {
				if (collide_target(i, quadFoundRedFlag - 1)) {
					remove_target(i);
					score++;
				}
			}
		}

		//Move targets and detect explosion
		for (i = num_targets - 1; i >= 0; i--) {
			targets[i][1] += targets[i][2];

			if (targets[i][1] >= HEIGHT) {
				remove_target(i);
				num_lives--;		//Lose a life
			}
		}

		//finally draw all the game elements
		draw_lives();
		draw_targets();
		draw_score();
	}
}
