/*
 * TestApp.c
 *
 * Uses:
 * 	CheckRedFlag.h
 * 	Game.h
 * 	Text.h
 * 
 * Description:
 * 	Initializes ZBT mux and red flag detection.
 * 	Runs the main game loop.
 *	Also provides base drawing functionality.
 *
 * Authors:
 * 	Leo Hwang
 * 	Tim Li
 */

// Located in: microblaze_0/include/xparameters.h
#include "xparameters.h"
#include "xutil.h"
#include "xbasic_types.h"
#include "xgpio.h"
#include "stdlib.h"

typedef unsigned char uchar;


//screen size
#define WIDTH	640
#define HEIGHT	480

//macros for screen writing
	//create a color value
long makecol (uchar r, uchar g, uchar b) {
	return ( (b) | ((g) << 8) | ((r) << 16) );
}

	//set a pixel to a color
#define putpixel(x, y, col)	\
	screen[y * WIDTH + x] = col;

//clear the screen (SLOW! b/c done in software)
void clear_screen (long col) {
	register long *screen = (long*)XPAR_ZBT_512KX32_MEM0_BASEADDR;
	register const long *screen_end = (long*)(screen + HEIGHT*WIDTH);

	while (screen < screen_end)
		*(screen++) = col;
}

//draw a rectangle
void rect (int x1, int y1, int x2, int y2, long col) {
	register const long *screen = (long*)XPAR_ZBT_512KX32_MEM0_BASEADDR;
	register long *pixel, *pixel_end;
	int y;

	for (y = y1; y < y2; y++)
	{
		pixel = (long*)(screen + y*WIDTH + x1);
		pixel_end = (long*)(pixel + x2 - x1);
		
		while (pixel < pixel_end)
			*(pixel++) = col;
	}
}


//LEAVE INCLUDES HERE, AS THEY DEPEND ON THE ABOVE FUNCTIONS!
#include "CheckRedFlag.h"
#include "Text.h"
#include "Game.h"
//#include "Matrix.h"


//global variables
XStatus status;
XGpio gpio_ZBTswitch;
int buffer_toggle;
int buffer_toggle_read;


//left-to-right scrolling rectangle at the bottom -- used to get some sense of the frame rate
#define WIDTH_GREEN	(5)
#define MAX_X_GREEN	(635)	//= 640-5
int x_green;


//update that green box
void doGreenBox()
{	
	//get a sense of frame speed
	rect(x_green, HEIGHT-WIDTH_GREEN, x_green+WIDTH_GREEN, HEIGHT, makecol(50, 255, 50)); //green
	if (++x_green == MAX_X_GREEN)
		x_green = 0;
}

//setup GPIO registers used to control screen buffer rotation
#define ZBTSWITCH_DIRMASK	(1)		//currently only using lowest bit
void SetupGPIO()
{
	//setup ZBT_Switch GPIO
	status = XGpio_Initialize(&gpio_ZBTswitch, XPAR_ZBT_SWITCH_CONTROL_DEVICE_ID);
	if (status != XST_SUCCESS)
		exit(0);

	//output on channel 1 (a 0 bit = output)
	XGpio_SetDataDirection(&gpio_ZBTswitch, 1, ~ZBTSWITCH_DIRMASK);
	//input on channel 2 (a 1 bit = input)
	XGpio_SetDataDirection(&gpio_ZBTswitch, 2, ZBTSWITCH_DIRMASK);
}

//send indication to rotate buffers, and hold until confirmed
void SwapBuffers()
{
	//need to toggle buffer bit to switch
	buffer_toggle = (buffer_toggle == 1) ? 0 : 1;
	XGpio_DiscreteWrite(&gpio_ZBTswitch, 1, buffer_toggle);

	//WAIT FOR SYNC WITH BLANKING PERIOD before drawing next frame
	while (buffer_toggle_read != buffer_toggle)
	{
		buffer_toggle_read = XGpio_DiscreteRead(&gpio_ZBTswitch, 2) & 1;
	}
}


int main (void)
{
	//setup variables
	int delay;
	int quadFoundRedFlag;

	buffer_toggle = 0;
	buffer_toggle_read = 0;
	x_green = 0;

	num_targets = 0;
	num_lives = 0;
	
	//setup hit-detection rectangles
	SetupQuadrants();
	
	//setup I/O registers
	SetupGPIO();
		
	while(1)
	{
		//DELAY for testing
		//for (delay=0; delay<1000000; delay++)
		//	;

		//swap buffers
		SwapBuffers();
		
		//find red area!
		quadFoundRedFlag = LocateFlagSector();

		//draw box marking quadrant found
		if (quadFoundRedFlag != 0)
			DrawBoxForQuadrant(quadFoundRedFlag);

		//scrolling green box at bottom of screen
		doGreenBox();

		//run game code
		GameProcFrame(quadFoundRedFlag);
		//MatrixProcFrame(quadFoundRedFlag);
	}
	
	return 0;
}

