//************************************************************************
//    This file is a modified version of the version provided by Xilinx
//    in the SVGA example
//    File: MEMORY_CTRL_532.v
//    Uses: ZBT_CONTROL_532.v
//    Used by: fsl_zbt_control_core.vhd
//    Modified for: University of Toronto ECE532 project
//                  FSL ZBT controller module
//    Modifications:
//      Used local defines instead of defines.v.
//      Removed an assign statement that was causing problems with ISE.
//      Instead, used signal concatenation in the declaration ({..,..,..})
//      
//    Below is the header provided by Xilinx:
//************************************************************************
//     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
//     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
//     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
//     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
//     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
//     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
//     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
//     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
//     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
//     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
//     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
//     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//     FOR A PARTICULAR PURPOSE.
//
//     (c) Copyright 2002 Xilinx, Inc.
//     All rights reserved.
//

/*
-------------------------------------------------------------------------------
   Title      : Memory Bank0 ZBT RAM Controller
   Project    : MicroBlaze Demo Board
-------------------------------------------------------------------------------
   File       : MEMORY_CTRL.v
   Company    : Xilinx, Inc.
   Created    : 2001/06/19
   Last Update: 2001/06/19
   Copyright  : (c) Xilinx Inc, 2001
-------------------------------------------------------------------------------
   Uses       : ZBT_CONTROL.v defines.v
-------------------------------------------------------------------------------
   Used by    : Any user logic that requires access to ZBT RAM
-------------------------------------------------------------------------------
   Description: This is the top level file for the ZBT memory controller interface
		  to the MEMORY  on the  MicroBlaze demo board.

		The design synchronously latched the address, data and control inputs 
		and where required adds pipeline delays before synchronously driving 
		the ZBT RAMs.

		When a read operation is requested the data is available FOUR clock 
		cycles later, one cycle latched the read request, two cycles are take 
		by the RAM and the fourth cycle to latch the data returned by the RAM.

		The control signals on the ZBT RAM are all derived from user_clocken,
		user_memory_read, user_memory_write, user_memory_ce,
		and the four byte write enables, which are all active HIGH.

		Two clocks are required; user_memory_clock, used by the ZBT 
		controller and user_memory_ram_clock, the clock that drives
		the ZBT RAM. These two clocks MUST be the same frequency and are
		sourced by the "CLOCKGEN" module. The user_memory_ram_clock
		uses a PCB trace for the DCM feedback path so that the clock on the
		ZBT RAM device is synchronized with the internal clock that drives 
		the ZBT controller.

		NOTE: The data bus is 32 bits wide, the parity bits are not supported,
			but individual byte writes are allowed.

			BYTE MAPPING

			BYTE_A <=> user_data[7:0]
			BYTE_B <=> user_data[15:8]
			BYTE_C <=> user_data[23:16]
			BYTE_D <=> user_data[31:24]


Conventions:
All EXTERNAL port signals are UPPER CASE.
Active HIGH PORT signals have a P suffix
Active LOW PORT signals have a N suffix.

All internal signals are LOWER CASE and are active HIGH.




-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
*/
module MEMORY_CTRL_532(
MEMORY_CLK_P,
MEMORY_CLKEN_N,
MEMORY_WEN_N,
MEMORY_WENA_N,
MEMORY_WENB_N,
MEMORY_WENC_N,
MEMORY_WEND_N,
MEMORY_ADV_LD_N,
MEMORY_OEN_N,
MEMORY_CEN_N,
MEMORY_ADDR_P,
MEMORY_DATA_A_P,
MEMORY_DATA_B_P,
MEMORY_DATA_C_P,
MEMORY_DATA_D_P,

user_memory_clock,
user_memory_ram_clock,
user_memory_clocken,
user_memory_load_addr,
user_memory_read,
user_memory_write,
user_memory_wea,
user_memory_web,
user_memory_wec,
user_memory_wed,
user_memory_ce,
user_memory_addr,
user_memory_read_data,
user_memory_write_data
);

`define ADDR_BITS 19 
`define DATA_BITS 32 

 

// INTERFACE TO ZBT RAM
output 			MEMORY_CLK_P;		// ZBT RAM clock
output 			MEMORY_CLKEN_N;	// ZBT RAM clock enable
output 			MEMORY_WEN_N;		// ZBT RAM write enable
output 			MEMORY_WENA_N;	// ZBT RAM byte A write enable
output 			MEMORY_WENB_N;	// ZBT RAM byte B write enable
output 			MEMORY_WENC_N;	// ZBT RAM byte C write enable
output 			MEMORY_WEND_N;	// ZBT RAM byte D write enable
output 			MEMORY_ADV_LD_N;	// ZBT RAM address load-advance
output 			MEMORY_OEN_N;		// ZBT RAM output enable
output 			MEMORY_CEN_N;		// ZBT RAM chip enable
output [(`ADDR_BITS-1):0]MEMORY_ADDR_P;// ZBT RAM address bus
inout  [7:0] 		MEMORY_DATA_A_P;	// ZBT RAM data bus byte "A"
inout  [7:0] 		MEMORY_DATA_B_P;	// ZBT RAM data bus byte "B"
inout  [7:0] 		MEMORY_DATA_C_P;	// ZBT RAM data bus byte "C"
inout  [7:0] 		MEMORY_DATA_D_P;	// ZBT RAM data bus byte "D"


// INTERFACE TO USER DESIGN
input 			user_memory_clock	;	// clock for control logic
input 			user_memory_ram_clock;	// clock for ZBT RAM
input 			user_memory_clocken;	// clock enable
input			user_memory_load_addr;	// flag to use the user_addr and NOT use the burst counter
input 			user_memory_read;		// flag to indicate a READ operation
input 			user_memory_write;		// flag to indicate a WRITE operation
input			user_memory_wea;		// byte "A" write enable
input			user_memory_web;		// byte "B" write enable
input			user_memory_wec;		// byte "C" write enable
input			user_memory_wed;		// byte "D" write enable
input 			user_memory_ce;		// flag to enable the memory opaeration
output [(`DATA_BITS-1):0]user_memory_read_data;	// data read FROM the memory
input  [(`DATA_BITS-1):0]user_memory_write_data;// data to be written TO the memory
input  [(`ADDR_BITS-1):0]user_memory_addr;	// address to be read or written




wire				ram_clock;
wire				ram_clocken_n;
wire				ram_we_n;
wire				ram_wea_n;
wire				ram_web_n;
wire				ram_wec_n;
wire				ram_wed_n;
wire				ram_oe_n;
wire				ram_ce0_n;
wire				ram_adv_ld;
wire [(`ADDR_BITS-1):0]	ram_addr;
wire [(`DATA_BITS-1):0]	ram_data_bus;
wire				user_ram_clock;
wire				user_clocken;
wire				user_read_writez;
wire				user_oe;
wire				user_ce0;
wire				user_ce1;
wire				user_ce2;
wire				user_adv_ld;
wire				user_wea;
wire				user_web;
wire				user_wec;
wire				user_wed;
wire				user_lbo;
wire				user_zz;
wire				clock;
wire [(`DATA_BITS-1):0]	user_write_data;
wire [(`DATA_BITS-1):0]	user_read_data;
wire [(`ADDR_BITS-1):0]	user_ram_addr;


wire [(`ADDR_BITS-1):0]	user_memory_addr;
wire 			user_memory_clock;
wire 			user_memory_ram_clock;
wire 			user_memory_clocken;
wire				user_memory_load_addr;
wire 			user_memory_read;
wire 			user_memory_write;
wire				user_memory_wea;
wire				user_memory_web;
wire				user_memory_wec;
wire				user_memory_wed;
wire				user_memory_oe;
wire 			user_memory_ce;
wire [(`DATA_BITS-1):0]	user_memory_read_data;
wire [(`DATA_BITS-1):0]	user_memory_write_data;

// make the connections between the user logic, memory controller and ZBT RAM
assign MEMORY_CLK_P		= ram_clock;
assign MEMORY_CLKEN_N		= ram_clocken_n;
assign MEMORY_WEN_N		= ram_we_n;
assign MEMORY_WENA_N		= ram_wea_n;
assign MEMORY_WENB_N		= ram_web_n;
assign MEMORY_WENC_N		= ram_wec_n;
assign MEMORY_WEND_N		= ram_wed_n;
assign MEMORY_ADV_LD_N		= ram_adv_ld;
assign MEMORY_OEN_N		= ram_oe_n;
assign MEMORY_CEN_N		= ram_ce0_n;

assign MEMORY_ADDR_P		= ram_addr;

assign user_memory_read_data = user_read_data;
assign user_ram_clock			= user_memory_ram_clock;
assign user_clocken				= user_memory_clocken;
assign user_read_writez			= !(!user_memory_read & user_memory_write);
assign user_oe					= (user_memory_ce & user_memory_read & !user_memory_write);
assign user_ce0				= (user_memory_ce & (user_memory_read ^ user_memory_write));
assign user_ce1				= 1'B1;	// CE1 is not used on the PCB
assign user_ce2				= 1'B1;	// CE2 is not used on the PCB
assign user_adv_ld				= user_memory_load_addr;
assign user_wea				= user_memory_wea;
assign user_web				= user_memory_web;
assign user_wec				= user_memory_wec;
assign user_wed				= user_memory_wed;
assign user_lbo				= 1'B0;	// BURST ORDER is set to linear on the PCB
assign user_zz					= 1'B0;	// SLEEP mode is disabled on the PCB
assign clock					= user_memory_clock;
assign user_write_data			= user_memory_write_data;
assign user_ram_addr			= user_memory_addr;

// instantiate the ZBT memory controller
ZBT_CONTROL_532 ZBT_CONTROL_532(
ram_clock,
ram_clocken_n,
ram_we_n,
ram_oe_n,
ram_ce0_n,
ram_ce1,
ram_ce2_n,
ram_adv_ld,
ram_wea_n,
ram_web_n,
ram_wec_n,
ram_wed_n,
ram_lbo,
ram_zz,
ram_addr,
user_read_data,
{MEMORY_DATA_D_P, MEMORY_DATA_C_P, MEMORY_DATA_B_P, MEMORY_DATA_A_P},
//ram_data_bus,
user_ram_clock,
user_clocken,
user_read_writez,
user_oe,
user_ce0,
user_ce1,
user_ce2,
user_adv_ld,
user_wea,
user_web,
user_wec,
user_wed,
user_lbo,
user_zz,
clock,
user_write_data,
user_ram_addr
);



endmodule //MEMORY_RAM_CTRL.v

