/*
 * File: tvin2.v
 * Uses: video_input_new.v
 *       memory_ctrl_532.v
 *
 * Description: 
 * 	tvin2 is the top level file of the module which transfers
 * 	video data into ZBT memory.
 * 
 * Author: Leo Hwang
 *
 */

module tvin2 (
// video_inputs ports
clk,
rst,
VIDEO_INPUT,
led0_o,
led1_o, // for debug

// memory_ctrl_532 ports
MEMORY_CLK_P,
MEMORY_CLKEN_N,
MEMORY_WEN_N,
MEMORY_WENA_N,
MEMORY_WENB_N,
MEMORY_WENC_N,
MEMORY_WEND_N,
MEMORY_ADV_LD_N,
MEMORY_OEN_N,
MEMORY_CEN_N,
MEMORY_ADDR_P,
MEMORY_DATA_A_P,
MEMORY_DATA_B_P,
MEMORY_DATA_C_P,
MEMORY_DATA_D_P
);

input clk;
input rst;
input [7:0] VIDEO_INPUT;	// Connect the MSB pins
output led0_o, led1_o;		// for debug

output MEMORY_CLK_P;		// ZBT RAM clock
output MEMORY_CLKEN_N;		// ZBT RAM clock enable
output MEMORY_WEN_N;		// ZBT RAM write enable
output MEMORY_WENA_N;		// ZBT RAM byte A write enable
output MEMORY_WENB_N;		// ZBT RAM byte B write enable
output MEMORY_WENC_N;		// ZBT RAM byte C write enable
output MEMORY_WEND_N;		// ZBT RAM byte D write enable
output MEMORY_ADV_LD_N;		// ZBT RAM address load-advance
output MEMORY_OEN_N;		// ZBT RAM output enable
output MEMORY_CEN_N;		// ZBT RAM chip enable
output [18:0]MEMORY_ADDR_P;	// ZBT RAM address bus
inout  [7:0] MEMORY_DATA_A_P;	// ZBT RAM data bus byte "A"
inout  [7:0] MEMORY_DATA_B_P;	// ZBT RAM data bus byte "B"
inout  [7:0] MEMORY_DATA_C_P;	// ZBT RAM data bus byte "C"
inout  [7:0] MEMORY_DATA_D_P;	// ZBT RAM data bus byte "D"

wire led0_o, led1_o;

wire user_memory_clock;
wire user_memory_ram_clock;
wire user_memory_clocken;
wire user_memory_load_addr;
wire user_memory_read;
reg user_memory_write;
wire user_memory_wea;
wire user_memory_web;
wire user_memory_wec;
wire user_memory_wed;
wire user_memory_ce;
reg [18:0]user_memory_addr;
wire [31:0]user_memory_read_data;
reg [31:0]user_memory_write_data;

wire [7:0] R_data;
wire [7:0] G_data;
wire [7:0] B_data;
wire Fmid;
wire Vmid;
wire Hmid;

reg [18:0] y;
reg [18:0] x;
reg [18:0] counter;

wire [31:0] rgbdata;

reg [5:0] Hmid_prev;
reg [2:0] pres_state, next_state;
reg [9:0] pixel_ct; // for debugging

reg delayer;

assign led0_o = 1;
assign led1_o = 0;
  
assign user_memory_clock = clk;
assign user_memory_ram_clock = clk;
assign user_memory_clocken = 1;
assign user_memory_load_addr = 1;
assign user_memory_read = 0;
assign user_memory_wea = 1;
assign user_memory_web = 1;
assign user_memory_wec = 1;
assign user_memory_wed = 1;
assign user_memory_ce = 1;

//Combine R, G, B values into a single 32 bit number.
assign rgbdata[31:24] = 8'b0;
assign rgbdata[23:16] = R_data;
assign rgbdata[15:8] = G_data;
assign rgbdata[7:0] = B_data;

//Width of screen.
parameter WIDTH = 640;
//NTSC video is greater than 640 wide, so skip a few of the pixels on the
//left side
parameter ADJUSTX = 20;

always @ (posedge clk) begin
	//Reset y position on V signal.
	//Skip a few of the top lines of the NTSC video.
	if (Vmid == 1) y <= -10;

	//Reset x position on H signal.
	if (Hmid == 1) x <= 0;

	if (Hmid == 0 && Vmid == 0) begin
		if (delayer == 1) begin
			delayer <= 0;
			user_memory_write <= (x >= ADJUSTX && x < (ADJUSTX+WIDTH)) ? 1 : 0;

			//Set ZBT address based on even or odd line.
			if (Fmid == 0) begin
				user_memory_addr <= y * 2 * WIDTH + ((WIDTH-1) - (x - ADJUSTX));
			end
			else begin
				user_memory_addr <= y * 2 * WIDTH + WIDTH + ((WIDTH-1) - (x - ADJUSTX));
			end
			user_memory_write_data <= rgbdata;

			x <= x + 1;

			//Increment y once finished writing WIDTH number of
			//pixels of the current line.
			if (x == (ADJUSTX+WIDTH+1)) y <= y + 1;
		end
		else begin
			//Skip one clock cycle, because video_input gives
			//a new pixel every two cycles.
			
			delayer <= 1;
			user_memory_write <= 0;
		end
	end
	else begin
		user_memory_write <= 0;
	end
end


memory_ctrl_532 memory_ctrl_532(
MEMORY_CLK_P,
MEMORY_CLKEN_N,
MEMORY_WEN_N,
MEMORY_WENA_N,
MEMORY_WENB_N,
MEMORY_WENC_N,
MEMORY_WEND_N,
MEMORY_ADV_LD_N,
MEMORY_OEN_N,
MEMORY_CEN_N,
MEMORY_ADDR_P,
MEMORY_DATA_A_P,
MEMORY_DATA_B_P,
MEMORY_DATA_C_P,
MEMORY_DATA_D_P,

user_memory_clock,
user_memory_ram_clock,
user_memory_clocken,
user_memory_load_addr,
user_memory_read,
user_memory_write,
user_memory_wea,
user_memory_web,
user_memory_wec,
user_memory_wed,
user_memory_ce,
user_memory_addr,
user_memory_read_data,
user_memory_write_data
);

video_input video_input(
  clk, rst,
  VIDEO_INPUT,
  R_data, 
  G_data, 
  B_data, 
  Fmid,
  Vmid,
  Hmid
);

endmodule
