//************************************************************************
//    This file is a modified version of the version provided by Xilinx
//    for video input processing on the Multimedia board
//    File: v422to444_v2mult_24t_532.v
//    Uses: 
//    Used by: input2rgb.v
//    Modified for: University of Toronto ECE532 project
//                  FSL video input module
//    Modifications:
//      Since we aren't going for high definition quality and XAPP294 says
//      that for VHS quality, the Cr and Cb values can be merely replicated
//      but 2-pixel linear interpolation works better.  This design uses
//      the 2-pixel linear approach
//      This design also assumes 8-bit video instead of 10-bit video
//
//    NOTE: this means that you can ignore the huge equation in the
//    comments below
//
//    Below is the header provided by Xilinx:
//************************************************************************
//-----------------------------------------------------------------
// v422to444_v2mult_24t.v - 422 to 444 Format Converter
//               Virtex-II Video Demo Board
//
//
//
//
//                  Author: Gregg C. Hawkes
//                  Senior Staff Applications Engineer
//
//                  Video Applications
//                  Advanced Products Division
//                  Xilinx, Inc.
//
//                  Copyright (c) 1999 Xilinx, Inc.
//                  All rights reserved
//
//                  Date:   Aug. 6, 2001
//                  For:    Video Demo Board
//
//                  RESTRICTED RIGHTS LEGEND
//
//      This software has not been published by the author, and 
//      has been disclosed to others for the purpose of enhancing 
//      and promoting design productivity in Xilinx products.
//
//      Therefore use, duplication or disclosure, now and in the 
//      future should give consideration to the productivity 
//      enhancements afforded the user of this code by the author's 
//      efforts.  Thank you for using our products !
//
// Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
//              WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
//              IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
//              A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
//
//
// 
// Revision:
//          Aug. 6, 2001     Creation
//
//
// Other modules instanced in this design:
//
//          none
/*

BRIEF DESCRIPTION

The process of 4:2:2 to 4:4:4 is simply creating the missing Cr and Cb
components. This version accomplishes this task by merely duplicating
the Cr and Cb information.

DETAILED DESCRIPTION

The video standard ITU-R BT.601 was introduced as the need for
transporting digital component video between countries and standards
increased. The analog component R'G'B' can be sampled in a very regular
way and converted from 4:4:4 to the digital 4:2:2 format, essentially
cutting in half the number of different components, Cr and Cb. 

The digital data is efficiently stored or transmitted to a destination
that reverses the process, i.e. converts back to 4:4:4 format, and
produces analog YUV or R'G'B' for display.


422 TO 444 CONVERSION
---------------------


Bob Turney, Xilinx Labs supplied me with this one to try.


  -   4  = 18'h3FFFC
      6  = 18'h00006
  -  12  = 18'h3FFF4
     20  = 18'h00014
  -  32  = 18'h3FFE0
     48  = 18'h00030
  -  70  = 18'h3FFBA
    104  = 18'h00068
  - 152  = 18'h3FF68
    236  = 18'h000EC
  - 420  = 18'h3FE5C
   1300  = 18'h00514
   1300  = 18'h00514
  - 420  = 18'h3FE5C
    236  = 18'h000EC
  - 152  = 18'h3FF68
    104  = 18'h00068
  -  70  = 18'h3FFBA
     48  = 18'h00030
  -  32  = 18'h3FFE0
     20  = 18'h00014
  -  12  = 18'h3FFF4
      6  = 18'h00006
  -   4  = 18'h3FFFC


       CrCb[i] = (
             -    4*(CrCb[1]+CrCb[24])
             +    6*(CrCb[2]+CrCb[23])
             -   12*(CrCb[3]+CrCb[22])
             +   20*(CrCb[4]+CrCb[21])
             -   32*(CrCb[5]+CrCb[20])
             +   48*(CrCb[6]+CrCb[19])
             -   70*(CrCb[7]+CrCb[18])
             +  104*(CrCb[8]+CrCb[17])
             -  152*(CrCb[9]+CrCb[16])
             +  236*(CrCb[10]+CrCb[15])
             -  420*(CrCb[11]+CrCb[14])
             + 1300*(CrCb[12]+CrCb[13]))/2048; 


Design Information
------------------
xc2v1000-ff896-6

MULT18X18s:  12   30%
Slices      864   16%
FFs       1,224   11%
LUTs        409    3%
IO           49   11%
gates    62,199

Minimum period:   12.042ns (Maximum frequency: 83 MHz)
Maximum net delay:   5.124ns

Note: Using coregen the V2 Mults can be replaced with parallel, 10 bit
signed integer X 12 bit coefficients for roughly 68 LUTs and 34
Registers each. This alternative would save the high performance V2
Multipliers for other uses or allow the design to directly map to
SPARTAN families.

*/


`timescale 1ns / 100ps



module v422to444_v2mult_24t_532 (
rst,             // resets input data register and control
clk,             // video component rate clock, 27Mhz for SDTV

Fi,              // Low to High signals start of Field One
Vi,              // High signals Vertical Blanking
Hi,              // High signals Horizontal Blanking

Fo,              // Field signal delayed by pipe length
Vo,              // Vertical signal delayed by pipe length
Ho,              // Horizontal signal delayed by pipe length
ceo,             // output data rate is 1/2 the clock rate

YCrCb_in,        // video component data, I[8].F[2], twos complement

Y_out,           // Y out, I[8].F[2], twos complement, clamped
Cr_out,          // Cr out, I[8].F[2], twos complement, clamped
Cb_out          // Cb out, I[8].F[2], twos complement, clamped
);


input rst, clk, Hi, Vi, Fi;
input[7:0] YCrCb_in;

output [7:0]  Y_out, Cr_out, Cb_out;
wire [7:0]  Y_out, Cr_out, Cb_out;
output Ho, Vo, Fo, ceo;

reg [4:0] H_rg, V_rg, F_rg;
wire Ho, Vo, Fo, ceo, H_rising;

reg cnt, CrCb_count;
wire Y_ld;
reg [7:0] Cb_outsig, Cr_outsig;
//, select_real_CrCb, select_filt_CrCb;

reg [7:0] Y_rg, CrCb_rg;

// pixel component pipelines
reg [7:0] Y_pipe  [2:0];
reg [7:0] CrCb_pipe [3:0];


integer i;



//-----------------------------------------------------------------------
//
/*
Delay SMPTE control signals F, V, H, by an amount equivalent to the
modules pipe length. The will allow different modules to be swapped out
without changing exterior control. this occurs in most of my modules.
*/												    


always @ (posedge clk) begin
  if (rst) begin F_rg <= 0; V_rg <= 0; H_rg <= 0; end
  else begin
    F_rg[4:0] <= {F_rg[3:0], Fi};
    V_rg[4:0] <= {V_rg[3:0], Vi};
    H_rg[4:0] <= {H_rg[3:0], Hi};
  end
end

assign Ho = H_rg[4];
assign Vo = V_rg[4];
assign Fo = F_rg[4];



//-----------------------------------------------------------------------
//
// Identify Y and CrCb valid
//

assign H_rising = ~H_rg[0] & Hi;

always @ (posedge clk) begin
  if (rst | H_rising) cnt <= 1;
  else cnt <= ~cnt;
end

assign Y_ld = (cnt == 1) ? 1:0;
assign ceo = Y_ld;


//-----------------------------------------------------------------------
//
// Seperate the Y and CrCb data streams
//
always @ (posedge clk) begin
  if (rst) Y_rg <= 8'h10;
  else if (Y_ld)  Y_rg <= YCrCb_in;
  else Y_rg <= Y_rg;
end

always @ (posedge clk) begin
  if (rst) CrCb_rg <= 0;
  else if (~Y_ld) CrCb_rg <= YCrCb_in;
  else CrCb_rg <= CrCb_rg;
end



//-----------------------------------------------------------------------
//
// 10 bit Y, Cr and Cb pipe registers, connected head to tail.  There are 
// 12 Y registers, 48 CrCb registers.
//
always @ (posedge clk) begin
  if (rst) Y_pipe[0] <= 8'h10;
  else if (Y_ld)  Y_pipe[0] <= Y_rg;
  else Y_pipe[0] <= Y_pipe[0];
end

always @ (posedge clk) begin
  for (i = 1; i <= 2; i = i+1) begin
    if (rst) Y_pipe[i] <= 8'h10;
    else if (Y_ld) Y_pipe[i] <= Y_pipe[i-1];
    else Y_pipe[i] <= Y_pipe[i];
  end
end

always @ (posedge clk) begin
  if (rst) begin
    CrCb_count <= 1;
    CrCb_pipe[0] <= 0; 
  end
  else if (H_rising) begin
    CrCb_count <= 1;
    CrCb_pipe[0] <= 0; 
  end
  else if (~Y_ld) begin
    CrCb_pipe[0] <= CrCb_rg;
    CrCb_count <= ~CrCb_count; 
  end
  else begin
	CrCb_count <= CrCb_count; 
  	CrCb_pipe[0] <= CrCb_pipe[0];
  end
end

always @ (posedge clk) begin
  for (i = 1; i <= 3; i = i+1) begin
    if (rst) CrCb_pipe[i] <= 0;
    else if (~Y_ld) CrCb_pipe[i] <= CrCb_pipe[i-1];
    else CrCb_pipe[i] <= CrCb_pipe[i];
  end
end

assign Y_out = Y_pipe[2];
assign Cb_out = Cb_outsig;
assign Cr_out = Cr_outsig;

wire [7:0] CrCbv1, CrCbv2, CrCbv3, CrCbv4;
assign CrCbv1[7:0] = {1'b0,CrCb_pipe[0][7:1]};
assign CrCbv2[7:0] = {1'b0,CrCb_pipe[2][7:1]};
assign CrCbv3[7:0] = {1'b0,CrCb_pipe[1][7:1]};
assign CrCbv4[7:0] = {1'b0,CrCb_pipe[3][7:1]};

always @ (posedge clk) begin
  if (rst) begin
    Cb_outsig <= 0;
    Cr_outsig <= 0; 
  end
  else if (Y_ld && ~CrCb_count) begin
    // Need to rightshift to div by 2 (assuming msb...lsb)
    Cb_outsig <= CrCbv1 + CrCbv2;
    Cr_outsig <= CrCbv3 + CrCbv4;
  end
  else if (Y_ld && CrCb_count) begin
    // Need to rightshift to div by 2 (assuming msb...lsb)
    Cb_outsig <= CrCbv3 + CrCbv4;
    Cr_outsig <= CrCbv1 + CrCbv2; 
  end
end


endmodule






