//************************************************************************
//    This file was written for University of Toronto, ECE532 
//    File: addr_gen.v
//    Uses: None
//    Used by: fsl_video_input.v
//    Works with ProjectNavigator 6.2.03i
//    Description: Module to generate address for writing line of video 
//                 into frame buffer.  Assumes display module wants line
//                 length of 640 pixels.  Change `even_reset and `line_l
//                 otherwise.
//    NCKC August 2004
//************************************************************************ 

module addr_gen (
  clk,
  rst,
  Fin,
  Vin,
  Hin,
  Addr
  );

input clk, rst;
input Fin, Vin, Hin;
output [18:0] Addr;

wire clk, rst;
wire Fin, Vin, Hin;
wire [18:0] Addr;

reg [18:0] Addr_reg;

reg [2:0] pres_state;
reg [2:0] next_state;

reg [1:0] Hprev;


// For State Machine
parameter st_reset=0;
parameter st_wait=1;
parameter st_odd=2;
parameter st_blankA=3;
parameter st_blankB=4;
parameter st_even=5;

`define odd_reset 0
`define even_reset 640
`define line_l 640

assign Addr = Addr_reg;

always @ (posedge clk or posedge rst) begin
  if ( rst == 1'b1)
    pres_state <= st_reset;
  else 
    pres_state <= next_state;
end

// From Xilinx XAPP286 Dec. 2001, for NTSC
// Line #   F  V  H(EAV)  H(SAV)  Notes
// 1-3      1  1  1       0       BlankingB
// 4-19     0  1  1       0       BlankingA
// 20-263   0  0  1       0       Field 1 (Odd) Active video 244 Lines
// 264-265  0  1  1       0       BlankingA
// 266-282  1  1  1       0       BlankingB
// 283-525  1  0  1       0       Field 2 (Even) Active video 243 Lines
always @ (negedge clk or posedge rst) begin
  if ( rst == 1'b1)
    next_state <= st_reset;
  else begin
    case (pres_state)
      st_reset:
        begin
          if ((Fin == 1'b0) && (Vin == 1'b1))
            next_state <= st_wait;
          else
            next_state <= st_reset;
        end		
      st_wait: // F0, V1
        begin
          if ((Vin == 1'b0) && (Fin == 1'b0))
	    next_state <= st_odd;
          else if ((Vin == 1'b1)  && (Fin == 1'b1))
	    next_state <= st_reset;
	  else //Vin = 1 and Fin = 0: wait for vertical blanking to finish
	    next_state <= st_wait;
        end 
      st_odd:  // F0, V0
        begin
          if (Vin == 1'b1)
	    next_state <= st_blankA;
          else
	    next_state <= st_odd;
        end 
      st_blankA:  // F0, V1
        begin
          if (Vin == 1'b0)
	    next_state <= st_odd;
          else if (Fin == 1'b1)
	    next_state <= st_blankB;
          else 
	    next_state <= st_blankA; 
        end 
      st_blankB:  // F1, V1
        begin
          if (Vin == 1'b0)
	    next_state <= st_even;
          else if (Fin == 1'b0)
	    next_state <= st_blankA;
          else 
	    next_state <= st_blankB;
        end 
      st_even:  // F1, V0
        begin
          if (Vin == 1'b1)
	    next_state <= st_blankB;
          else
	    next_state <= st_even;
        end 
    endcase
  end
end

always @ (posedge clk or posedge rst) begin
  if ( rst == 1'b1)
    Hprev <= 1'b1;
  else 
    Hprev[0] <= Hin;
end


always @ (posedge clk or posedge rst) begin
  if ( rst == 1'b1)
    Addr_reg <= 19'h0;
  else begin
    case (pres_state)
      st_reset:
        begin
	  Addr_reg <= 19'h0;
        end		
      st_wait: // F0, V1
        begin
	  Addr_reg <= `odd_reset;
        end 
      st_odd:  // F0, V0
        begin
	  if (( Hprev == 1'b0 ) && ( Hin == 1'b1))
            Addr_reg <= Addr_reg + (2* `line_l);
          else
	    Addr_reg <= Addr_reg;
        end 
      st_blankA:  // F0, V1
        begin
	  Addr_reg <= `odd_reset;
        end 
      st_blankB:  // F1, V1
        begin
	  Addr_reg <= `even_reset;
        end 
      st_even:  // F1, V0
        begin
	     if (( Hprev == 1'b0 ) && ( Hin == 1'b1))
          Addr_reg <= Addr_reg + (2* `line_l);
        else
	       Addr_reg <= Addr_reg;
		 end
    endcase
  end
end

endmodule
