//************************************************************************
//    This file was written for University of Toronto, ECE532 
//    File: fsl_dual_zbt_control.v
//    Uses: fsl_zbt_control_core.v or fsl_zbt_control_core_halffull.v
//    Used by: top_dualzbt_dow.v
//    Works with ProjectNavigator 6.2.03i
//    Description: Module to wrap zbt controllers for use with 2 pairs of 
//    FSL connections This version of the ZBT controller is for the case 
//    when synchronization is critical for both devices attached to the
//    ZBT controller
//
//    Protocol for communication is:
//      control word:  upper 19 bits give # of words to be transmitted
//                     20th bit is read/write_n by peripheral
//      address word:  upper 19 bits give address (19 bits to address into
//                      a ZBT)
//      data:  must transmit exactly the number of data words you specified
//              you would in the control word, otherwise the FSLs get clogged
//
//    NCKC August 2004
//************************************************************************ 
//
// Make sure Logic_Clk is running at same frequency as user_memory_ram_clock
// The RAM clock should be generated by the DCM with the feedback loop that
// is connected to the feedback pins in the FPGA.  See CLOCKGEN_532 module.


module fsl_dual_zbt_control (
    Can_switch_n, // For debug to drive LED, 0 => Lit = can switch, 1 => can't
    FSL_Clk,
    Logic_Clk,
    Reset,

    // "Batch" signals
    FSL0_batch,  // 1 => batch in progress, don't swap memories. Be careful
    FSL1_batch,  //      b/c want some opportunity to swap almost each batch

    // Slave FSL0 Signals
    FSL0_S_Clk,
    FSL0_S_Read,
    FSL0_S_Data,
    FSL0_S_Control,
    FSL0_S_Exists,
    // Master FSL0 Signals
    FSL0_M_Clk,
    FSL0_M_Write,
    FSL0_M_Data,
    FSL0_M_Control,
    FSL0_M_Full,
    // Slave FSL1 Signals
    FSL1_S_Clk,
    FSL1_S_Read,
    FSL1_S_Data,
    FSL1_S_Control,
    FSL1_S_Exists ,
    // Master FSL1 Signals
    FSL1_M_Clk,
    FSL1_M_Write,
    FSL1_M_Data,
    FSL1_M_Control,
    FSL1_M_Full,

    // ZBT0 signals
    MEMORY0_CLK_P,
    MEMORY0_CLKEN_N,
    MEMORY0_WEN_N,
    MEMORY0_WENA_N,
    MEMORY0_WENB_N,
    MEMORY0_WENC_N,
    MEMORY0_WEND_N,
    MEMORY0_ADV_LD_N,
    MEMORY0_OEN_N,
    MEMORY0_CEN_N,
    MEMORY0_ADDR_P,
    MEMORY0_DATA_A_P,
    MEMORY0_DATA_B_P,
    MEMORY0_DATA_C_P,
    MEMORY0_DATA_D_P,
    
    // ZBT1 signals
    MEMORY1_CLK_P,
    MEMORY1_CLKEN_N,
    MEMORY1_WEN_N,
    MEMORY1_WENA_N,
    MEMORY1_WENB_N,
    MEMORY1_WENC_N,
    MEMORY1_WEND_N,
    MEMORY1_ADV_LD_N,
    MEMORY1_OEN_N,
    MEMORY1_CEN_N,
    MEMORY1_ADDR_P,
    MEMORY1_DATA_A_P,
    MEMORY1_DATA_B_P,
    MEMORY1_DATA_C_P,
    MEMORY1_DATA_D_P,
    
    user_memory_ram_clock_zbt,
    user_memory_clocken_zbt,
    user_memory_load_addr_zbt,

    user_memory_ce_zbt
);


output Can_switch_n; // For debug
input FSL_Clk;
input Logic_Clk;
input Reset;

input FSL0_batch;
input FSL1_batch;

// Slave FSL0 Signals
output FSL0_S_Clk;
output FSL0_S_Read;
input [0:31] FSL0_S_Data;
input FSL0_S_Control;
input FSL0_S_Exists;

// Master FSL0 Signals
output FSL0_M_Clk;
output FSL0_M_Write;
output [0:31] FSL0_M_Data;
output FSL0_M_Control;
input FSL0_M_Full;

// Slave FSL1 Signals
output FSL1_S_Clk;
output FSL1_S_Read;
input [0:31] FSL1_S_Data;
input FSL1_S_Control;
input FSL1_S_Exists ;

// Master FSL1 Signals
output FSL1_M_Clk;
output FSL1_M_Write;
output [0:31] FSL1_M_Data;
output FSL1_M_Control;
input FSL1_M_Full;

// ZBT0 signals
output MEMORY0_CLK_P;
output MEMORY0_CLKEN_N;
output MEMORY0_WEN_N;
output MEMORY0_WENA_N;
output MEMORY0_WENB_N;
output MEMORY0_WENC_N;
output MEMORY0_WEND_N;
output MEMORY0_ADV_LD_N;
output MEMORY0_OEN_N;
output MEMORY0_CEN_N;
output [18:0] MEMORY0_ADDR_P;
inout [7:0] MEMORY0_DATA_A_P;
inout [7:0] MEMORY0_DATA_B_P;
inout [7:0] MEMORY0_DATA_C_P;
inout [7:0] MEMORY0_DATA_D_P;
 
// ZBT1 signals
output MEMORY1_CLK_P;
output MEMORY1_CLKEN_N;
output MEMORY1_WEN_N;
output MEMORY1_WENA_N;
output MEMORY1_WENB_N;
output MEMORY1_WENC_N;
output MEMORY1_WEND_N;
output MEMORY1_ADV_LD_N;
output MEMORY1_OEN_N;
output MEMORY1_CEN_N;
output [18:0] MEMORY1_ADDR_P;
inout [7:0] MEMORY1_DATA_A_P;
inout [7:0] MEMORY1_DATA_B_P;
inout [7:0] MEMORY1_DATA_C_P;
inout [7:0] MEMORY1_DATA_D_P;
 
input user_memory_ram_clock_zbt;
input user_memory_clocken_zbt;
input user_memory_load_addr_zbt;

input user_memory_ce_zbt;



reg Can_switch_n; // For debug
wire FSL_Clk;
wire Logic_Clk;
wire Reset;

wire FSL0_batch;
wire FSL1_batch;

// Slave FSL0 Signals
wire FSL0_S_Clk;
wire FSL0_S_Read;
wire [0:31] FSL0_S_Data;
wire FSL0_S_Control;
wire FSL0_S_Exists;

// Master FSL0 Signals
wire FSL0_M_Clk;
wire FSL0_M_Write;
wire [0:31] FSL0_M_Data;
wire FSL0_M_Control;
wire FSL0_M_Full;

// Slave FSL1 Signals
wire FSL1_S_Clk;
wire FSL1_S_Read;
wire [0:31] FSL1_S_Data;
wire FSL1_S_Control;
wire FSL1_S_Exists ;

// Master FSL1 Signals
wire FSL1_M_Clk;
wire FSL1_M_Write;
wire [0:31] FSL1_M_Data;
wire FSL1_M_Control;
wire FSL1_M_Full;

// ZBT0 signals
wire MEMORY0_CLK_P;
wire MEMORY0_CLKEN_N;
wire MEMORY0_WEN_N;
wire MEMORY0_WENA_N;
wire MEMORY0_WENB_N;
wire MEMORY0_WENC_N;
wire MEMORY0_WEND_N;
wire MEMORY0_ADV_LD_N;
wire MEMORY0_OEN_N;
wire MEMORY0_CEN_N;
wire [18:0] MEMORY0_ADDR_P;
wire [7:0] MEMORY0_DATA_A_P;
wire [7:0] MEMORY0_DATA_B_P;
wire [7:0] MEMORY0_DATA_C_P;
wire [7:0] MEMORY0_DATA_D_P;
 
// ZBT1 signals
wire MEMORY1_CLK_P;
wire MEMORY1_CLKEN_N;
wire MEMORY1_WEN_N;
wire MEMORY1_WENA_N;
wire MEMORY1_WENB_N;
wire MEMORY1_WENC_N;
wire MEMORY1_WEND_N;
wire MEMORY1_ADV_LD_N;
wire MEMORY1_OEN_N;
wire MEMORY1_CEN_N;
wire [18:0] MEMORY1_ADDR_P;
wire [7:0] MEMORY1_DATA_A_P;
wire [7:0] MEMORY1_DATA_B_P;
wire [7:0] MEMORY1_DATA_C_P;
wire [7:0] MEMORY1_DATA_D_P;
 
wire user_memory_ram_clock_zbt;
wire user_memory_clocken_zbt;
wire user_memory_load_addr_zbt;

wire user_memory_ce_zbt;

// Internal Signals
reg FSL_select; // If 0 then FSL0 talks to C0, FSL1 to C1
                // If 1 then FSL1 talks to C0, FSL0 to C1
reg [1:0] delsw;
wire Can_Switch_FSLs, Can_Switch_FSLs_C0, Can_Switch_FSLs_C1;
wire delok;
wire ctrl_exist;

wire FSL_S_Read_C0, FSL_S_Read_C1;
wire FSL_S_Control_C0, FSL_S_Control_C1;
wire [0:31] FSL_S_Data_C0, FSL_S_Data_C1;
wire FSL_S_Exists_C0, FSL_S_Exists_C1;
wire FSL_M_Control_C0, FSL_M_Control_C1;
wire [0:31] FSL_M_Data_C0, FSL_M_Data_C1;
wire FSL_M_Write_C0, FSL_M_Write_C1;
wire FSL_M_Full_C0, FSL_M_Full_C1;


assign delok = !(delsw[0] | delsw[1]);
assign ctrl_exist = (FSL_S_Control_C0 & FSL_S_Exists_C0) | (FSL_S_Control_C1 & FSL_S_Exists_C1);


assign Can_Switch_FSLs = Can_Switch_FSLs_C0 & Can_Switch_FSLs_C1 & (!FSL0_batch) & (!FSL1_batch);

assign FSL0_S_Clk = FSL_Clk;
assign FSL1_S_Clk = FSL_Clk;
assign FSL0_M_Clk = FSL_Clk;
assign FSL1_M_Clk = FSL_Clk;

// Does doing the MUX logic asynchronously cause potential timing problems?
assign FSL0_S_Read = ((FSL_select == 1'b0) ? FSL_S_Read_C0 : FSL_S_Read_C1);
assign FSL1_S_Read = ((FSL_select == 1'b0) ? FSL_S_Read_C1 : FSL_S_Read_C0);

assign FSL_S_Data_C0 = ((FSL_select == 1'b0) ? FSL0_S_Data : FSL1_S_Data);
assign FSL_S_Data_C1 = ((FSL_select == 1'b0) ? FSL1_S_Data : FSL0_S_Data);

assign FSL_S_Control_C0 = ((FSL_select == 1'b0) ? FSL0_S_Control : FSL1_S_Control);
assign FSL_S_Control_C1 = ((FSL_select == 1'b0) ? FSL1_S_Control : FSL0_S_Control);

assign FSL_S_Exists_C0 = ((FSL_select == 1'b0) ? FSL0_S_Exists : FSL1_S_Exists);
assign FSL_S_Exists_C1 = ((FSL_select == 1'b0) ? FSL1_S_Exists : FSL0_S_Exists);

assign FSL0_M_Write = ((FSL_select == 1'b0) ? FSL_M_Write_C0 : FSL_M_Write_C1);
assign FSL1_M_Write = ((FSL_select == 1'b0) ? FSL_M_Write_C1 : FSL_M_Write_C0);

assign FSL0_M_Data = ((FSL_select == 1'b0) ? FSL_M_Data_C0 : FSL_M_Data_C1);
assign FSL1_M_Data = ((FSL_select == 1'b0) ? FSL_M_Data_C1 : FSL_M_Data_C0);

assign FSL0_M_Control = ((FSL_select == 1'b0) ? FSL_M_Control_C0 : FSL_M_Control_C1);
assign FSL1_M_Control = ((FSL_select == 1'b0) ? FSL_M_Control_C1 : FSL_M_Control_C0);

assign FSL_M_Full_C0 = ((FSL_select == 1'b0) ? FSL0_M_Full : FSL1_M_Full);
assign FSL_M_Full_C1 = ((FSL_select == 1'b0) ? FSL1_M_Full : FSL0_M_Full);



// Select FSL Muxing configuration
always @ (posedge Logic_Clk or posedge Reset)
begin
  if (Reset == 1'b1) begin
    FSL_select <= 1'b0;
    delsw <= 2'b00;
    Can_switch_n <= 1'b0;
  end
  else begin
    delsw[1] <= delsw[0];

    if ((Can_Switch_FSLs == 1'b1) && (delok == 1'b1)) begin
    // Didn't switch in the last 2 cycles already
      Can_switch_n <= ((ctrl_exist == 1'b1) ? 1'b1 : 1'b0); // active low LED
      FSL_select <= !FSL_select;
      delsw[0] <= 1'b1;
    end
    else begin
      Can_switch_n <= 1'b0;
      FSL_select <= FSL_select;
      delsw[0] <= 1'b0;
    end
  end
end
	
// Do I have to worry about the VHDL generics?
fsl_zbt_control_core fsl_zbt_ctrl_core0 (
    .FSL_Clk(FSL_Clk), // common
    .Logic_Clk(Logic_Clk), // common
    .Reset(Reset), // common
    .Can_Switch_FSLs_o(Can_Switch_FSLs_C0), 
    .FSL_S_Read_o(FSL_S_Read_C0), 
    .FSL_S_Data_i(FSL_S_Data_C0), 
    .FSL_S_Control_i(FSL_S_Control_C0), 
    .FSL_S_Exists_i(FSL_S_Exists_C0), 
    .FSL_M_Write_o(FSL_M_Write_C0), 
    .FSL_M_Data_o(FSL_M_Data_C0), 
    .FSL_M_Control_o(FSL_M_Control_C0), 
    .FSL_M_Full_i(FSL_M_Full_C0), 
    .MEMORY0_CLK_P_o(MEMORY0_CLK_P), 
    .MEMORY0_CLKEN_N_o(MEMORY0_CLKEN_N), 
    .MEMORY0_WEN_N_o(MEMORY0_WEN_N), 
    .MEMORY0_WENA_N_o(MEMORY0_WENA_N), 
    .MEMORY0_WENB_N_o(MEMORY0_WENB_N), 
    .MEMORY0_WENC_N_o(MEMORY0_WENC_N), 
    .MEMORY0_WEND_N_o(MEMORY0_WEND_N), 
    .MEMORY0_ADV_LD_N_o(MEMORY0_ADV_LD_N), 
    .MEMORY0_OEN_N_o(MEMORY0_OEN_N), 
    .MEMORY0_CEN_N_o(MEMORY0_CEN_N), 
    .MEMORY0_ADDR_P_o(MEMORY0_ADDR_P), 
    .MEMORY0_DATA_A_P_io(MEMORY0_DATA_A_P), 
    .MEMORY0_DATA_B_P_io(MEMORY0_DATA_B_P), 
    .MEMORY0_DATA_C_P_io(MEMORY0_DATA_C_P), 
    .MEMORY0_DATA_D_P_io(MEMORY0_DATA_D_P), 
    .user_memory_ram_clock_zbt_i(user_memory_ram_clock_zbt), // common
    .user_memory_clocken_zbt_i(user_memory_clocken_zbt), // common
    .user_memory_load_addr_zbt_i(user_memory_load_addr_zbt), // common
    .user_memory_ce_zbt_i(user_memory_ce_zbt)  // common
    );


fsl_zbt_control_core fsl_zbt_ctrl_core1 (
    .FSL_Clk(FSL_Clk), // common
    .Logic_Clk(Logic_Clk), // common
    .Reset(Reset), // common
    .Can_Switch_FSLs_o(Can_Switch_FSLs_C1), 
    .FSL_S_Read_o(FSL_S_Read_C1), 
    .FSL_S_Data_i(FSL_S_Data_C1), 
    .FSL_S_Control_i(FSL_S_Control_C1), 
    .FSL_S_Exists_i(FSL_S_Exists_C1), 
    .FSL_M_Write_o(FSL_M_Write_C1), 
    .FSL_M_Data_o(FSL_M_Data_C1), 
    .FSL_M_Control_o(FSL_M_Control_C1), 
    .FSL_M_Full_i(FSL_M_Full_C1), 
    .MEMORY0_CLK_P_o(MEMORY1_CLK_P), 
    .MEMORY0_CLKEN_N_o(MEMORY1_CLKEN_N), 
    .MEMORY0_WEN_N_o(MEMORY1_WEN_N), 
    .MEMORY0_WENA_N_o(MEMORY1_WENA_N), 
    .MEMORY0_WENB_N_o(MEMORY1_WENB_N), 
    .MEMORY0_WENC_N_o(MEMORY1_WENC_N), 
    .MEMORY0_WEND_N_o(MEMORY1_WEND_N), 
    .MEMORY0_ADV_LD_N_o(MEMORY1_ADV_LD_N), 
    .MEMORY0_OEN_N_o(MEMORY1_OEN_N), 
    .MEMORY0_CEN_N_o(MEMORY1_CEN_N), 
    .MEMORY0_ADDR_P_o(MEMORY1_ADDR_P), 
    .MEMORY0_DATA_A_P_io(MEMORY1_DATA_A_P), 
    .MEMORY0_DATA_B_P_io(MEMORY1_DATA_B_P), 
    .MEMORY0_DATA_C_P_io(MEMORY1_DATA_C_P), 
    .MEMORY0_DATA_D_P_io(MEMORY1_DATA_D_P), 
    .user_memory_ram_clock_zbt_i(user_memory_ram_clock_zbt), // common
    .user_memory_clocken_zbt_i(user_memory_clocken_zbt), // common
    .user_memory_load_addr_zbt_i(user_memory_load_addr_zbt), // common
    .user_memory_ce_zbt_i(user_memory_ce_zbt)  // common
    );

endmodule
