//************************************************************************
//    File: fsl_mem_controller.v
//    Uses: mem_controller.v
//          arbiter.v
//          fsl_mem_program.v
//    Provides a programmable interface to shared memory
//

module fsl_mem_controller (
//Input Connections for controller:
  clk,
  rst,
//Connections to Memory Controller:
  //Outputs:
//  mem_clk, //y
 // mem_rst, //y
  mem_addr, //y
  mem_in_data, //y
  fifo_empty, //y
  fifo_full, //y
  read_not_write, //y
  enable, //y
  //Inputs:
  read_fifo, //y
  write_fifo, //y
  mem_out_data, //y
  running, //y
  done, //y

  //DEBUGGING:
  led0,
  led1,

//Connected to arbiter:
  addr_msb_req,  //Used to req the correct memory bank
  req_mem_arb,  //status request bit to arbiter
  control_mem0,  //status bit from arbiter
  control_mem1,  //status bit from arbiter
  addr_msb_ctrl,

//Connections to the FSL Memory Controller Program:
  prog_clk, //y
  prog_rst, //y
  prog_instr,
  program_cbit,
  prog_instr_read,
  valid_instr,
  //sbit_using_mem0,  //status bit indicating mem0 in use by controller
  //**mem0_done //***should this be generic "mem" from controller?NECESSARY?


  // Slave FSL Signals
  FSL_S_Clk, //y
  FSL_S_Read, //y
  FSL_S_Data, //y
  FSL_S_Control,
  FSL_S_Exists, //y

  // Master FSL Signals
  FSL_M_Clk, //y
  FSL_M_Write, //y
  FSL_M_Data, //y
  FSL_M_Control,
  FSL_M_Full //y
);
  
  input clk, rst;
  //output mem_clk, mem_rst; 
  output fifo_empty, fifo_full; 
  output read_not_write, enable;
  output [18:0] mem_addr;
  output [31:0] mem_in_data;
  input read_fifo, write_fifo, running, done;
  input [31:0] mem_out_data;

  //debugging:
  output led0, led1;

  //arbiter:
  output [12:0] addr_msb_req;
  output req_mem_arb;
  input control_mem0;
  input control_mem1;
  input [12:0] addr_msb_ctrl;

  
  output prog_clk, prog_rst;
  input [0:31] prog_instr;
  input program_cbit;
  output prog_instr_read;
  input valid_instr;
  //output sbit_using_mem0;


  // Slave FSL Signals
  output FSL_S_Clk;
  output FSL_S_Read;
  input  [0:31] FSL_S_Data;
  input  FSL_S_Control;
  input  FSL_S_Exists;

  // Master FSL Signals
  output FSL_M_Clk;
  output FSL_M_Write;
  output [0:31] FSL_M_Data;
  output FSL_M_Control;
  input FSL_M_Full;

  //Connections for inputs and outputs:
  wire clk, rst; 
  //wire mem_clk, mem_rst; 
  wire fifo_empty, fifo_full;
  wire read_not_write; 
  reg enable;
  reg [18:0] mem_addr;
  reg [18:0] base_addr;
  wire [31:0] mem_in_data;
  wire read_fifo, write_fifo, running, done;
  wire [31:0] mem_out_data;

  //debugging:
  wire led0, led1;

  reg [12:0] addr_msb_req;
  wire req_mem_arb;
  wire control_mem0;
  wire control_mem1;
  wire prog_clk, prog_rst, program_cbit;
  wire [0:31] prog_instr;
  wire prog_instr_read, valid_instr; //, sbit_using_mem0;


  wire FSL_S_Clk;
  wire FSL_S_Read;
  //reg FSL_S_Read_reg;
  wire [0:31] FSL_S_Data;
  wire FSL_S_Control;
  wire FSL_S_Exists;
  wire FSL_M_Clk;
  //reg FSL_M_Write;
  //reg [0:31] FSL_M_Data;
  wire FSL_M_Write;
  wire [0:31] FSL_M_Data;
  wire FSL_M_Control;
  wire FSL_M_Full;

//`define line_length 10'd640 

//Create constants for all the instruction opcode:
`define reset 8'h00
`define noop 8'h04

`define readR_Imm 8'h02
`define write_Imm 8'h03
`define readA_Imm 8'h01

`define readR_Abs_addr 8'h06  
`define write_Abs_addr 8'h07  // These first 8 instructions will be 
`define readA_Abs_addr 8'h05  // read and transmitted.

`define read_Areg 8'h0A
`define write_Areg 8'h0B

`define add_imm_Areg 8'h18
`define sub_imm_Areg 8'h08
`define wait_rx 8'h0C

`define read_Areg_indirect 8'h0E
`define write_Areg_indirect 8'h0F

`define read_Areg_autoinc 8'h1E
`define write_Areg_autoinc 8'h1F

//NEW INSTRUCTION TO ADD??
`define set_clear_status_bits 8'h13

//States for overall execution path state machine:
parameter get_next_instr=0;
parameter decode_instr=1;
parameter get_memory=2;
parameter execute_instr=3;
  
  reg new_mem_op;
  reg valid_msbs;
//state and delay signals:
  //reg [2:0] pres_state;
  //reg [2:0] next_state;
  reg [1:0] pres_state;
  reg [1:0] next_state;

 //IR signals:
  wire  rx_IR_full;
  wire [0:31] rx_IR;
  wire  rx_cbit;  //receive control bit
  wire prog_IR_full;
  wire [0:31] prog_IR;
  wire prog_cbit;  //program control bit
  reg  ex_cbit;  //executing control bit
  reg exec_prog; //am I executing the program or an rx_IR**only valid when 
  			//ex_IR_full is high
  wire ex_IR_full;

  //Address registers:
  reg [0:31] a0_reg; //Address register 0 
  reg [0:23] a0_operand; //latches the data portion of the instruction
  reg write_a0;
  wire [0:12] sel_imm_addr_msb; //Used to indicate which memory bank was used

  //Mux control signals:
  wire [0:31] a0_mux_output; //multiplexer output to the a0 reg
  reg [0:31] FSL_out_mux; //fifo_mux_out; //output from multiplexer into buffering fifo
  
  //FSL and FIFO control signals:
  reg fsl_write;
  //wire fifo_write; //attaches to either a buffering fifo or the output FSL
  //reg fifo_read; //used to read fsl data onto VGA display
  //wire fifo_read; //used to read fsl data onto VGA display
  //wire fifo_empty;
  //wire fifo_full;
  //wire fifo_exists;

  //Used to generate Computing element resets:
  wire control_reset;  //combines system reset and instruction reset
  reg instr_reset;	//generated via the reset instruction- clock period long
  //Used to interface the ex_IR, A0 and data with the fifo: 
  reg [1:0] mux_select_fifo_in; //used to select
  wire [0:7] ex_IR_fifo_in;
  reg [7:0] opcode;  //Used to store the opcode for 2-part instructions

  //Control registers:	
  reg change_a0;  //Flag to indicate when instruction will alter the value of a0
  reg IR_req_W0;  // Indicates when the instruction writes to the fifo
  reg IR_req_W1;  // Indicates when the instruction writes an immed to the fifo
  reg Addr_req_W;  // Indicates when the address register needs to be written to
  				//the fifo
  reg data_req_W;  // Indicates that there is data to be written to the fifo
  reg rx_data_req; // Used to indicate when data should be read in as part of
  			// the received instruction.
  reg rx_addr_req; // Used to indicate when an address is read in as part of
  			// the received instruction.
  reg req_mem;
  reg [0:31] new_a0_total;  //Output from Accumulator;
  reg [23:0] data_cntr; // Used to keep track of the number of reads/writes
  				// to memory

  //Program interface and Instruction Register Read signals:
  //wire [0:31] prog_instr;  //connects to the program instrucion input port
  //wire program_cbit; //connects to the program's control bit
  //wire prog_instr_read; //reading from the port into the program IR
  wire [0:31] ex_IR;  //wire used as mux output from prog and rx IRs
  wire ex_cbit_input; //wire used as mux output from prog and rx cbits
  wire ex_instr_read;  //reads an instruction into the ex_IR
  wire reading_prog_IR;  //reading the prog_IR clears it
  wire reading_rx_IR;  //reading the rx_IR clears it
  wire sel_ex_IR_input; // selects the input to the ex_IR
  //wire rx_req_W;
  //wire select_mem_addr_in;  //Status bit 1- send to program

  //Status bit:
  //wire sbit_using_mem0;  //Status bit 0- latch video_write on this signal
 
  //reg [11:0] memdata_ctr; // Max pix are 1024, but don't want overflow errors


//assign select_mem_addr_in = opcode[3] ? a0_reg[18:0]:
//These are the easy connections:  

//debugging:
reg invalid_opcode;

always @ (posedge clk or posedge rst) begin 
  if (rst == 1'b1)
    invalid_opcode <= 1'b0;
  else if ((pres_state == execute_instr) && ((opcode == 8'h00) ||
	  (opcode == 8'h01) || (opcode == 8'h02)))
    invalid_opcode <= 1'b1;
  else
    invalid_opcode <= invalid_opcode;
end
assign led0 = (invalid_opcode == 1'b1) ? 1'b0 : 1'b1;
assign led1 = (rst == 1'b1) ? 1'b0 : 1'b1;
//assign led1 = (pres_state == execute_instr) ? 1'b0 : 1'b1;

//assign mem_clk = clk;
assign FSL_S_Clk = clk;
assign FSL_M_Clk = clk;
//assign mem_rst = control_reset;
assign FSL_S_Read = ((read_fifo == 1'b1) || (reading_rx_IR == 1'b1));
//assign FSL_S_Read = ((reading_rx_IR == 1'b1) || (read_fifo == 1'b1));
//assign FSL_M_Write = ((write_fifo == 1'b1) || (fsl_write == 1'b1);
assign FSL_M_Write = ((fsl_write == 1'b1)); 
assign mem_in_data = FSL_S_Data;
//assign mem_out_data = FSL_M_Data;
assign fifo_empty = ~FSL_S_Exists;
assign fifo_full = FSL_M_Full;
assign read_not_write = sel_ex_IR_input ? ~opcode[0] : opcode[0];
assign FSL_M_Control = (mux_select_fifo_in == 2'b00) ? IR_req_W0 : 1'b0;
assign sel_imm_addr_msb = (exec_prog == 1'b1) ? 
			prog_instr[0:12] : rx_IR[0:12] ;

assign req_mem_arb = valid_msbs && req_mem;
  
//assign sbit_using_mem0 = control_mem0;

//NEEDED WHEN THE PROGRAM IS OUTSIDE OF THIS CONTROLLER:
assign prog_IR_full = valid_instr;
assign prog_clk = clk;
assign prog_rst = control_reset;


//assign fifo_write = ((pres_state == execute_instr) && (rx_addr_req == 1'b0)
			//&& (rx_data_req == 1'b1) && (fifo_full == 1'b0));

assign FSL_M_Data = FSL_out_mux;

//assign fifo_read = ((valid_data == 1'b1) && (fifo_empty == 1'b0));

//// make the internal connections
//// Should I be registering FSL_S_Data?

assign ex_IR_fifo_in[0:4] = 5'b00000;
assign ex_IR_fifo_in[5] = opcode[2];
assign ex_IR_fifo_in[6:7] = (exec_prog == 1'b1) ? opcode[1:0] : ~opcode[1:0];

//assign valid_data = ((valid_data_reg == 1'b1) && (data_req_W == 1'b1));
assign ex_instr_read = ((pres_state == get_next_instr) 
	 && ((rx_IR_full || prog_IR_full)));  
//Used to mux the outputs from the two different IRs:
assign sel_ex_IR_input = (pres_state == get_next_instr) ? 
				rx_IR_full : ~exec_prog;
assign ex_IR = sel_ex_IR_input ? rx_IR : prog_IR ;
assign ex_IR_full = sel_ex_IR_input ? rx_IR_full : prog_IR_full;
assign ex_cbit_input = sel_ex_IR_input ? rx_cbit : prog_cbit ;
   
assign prog_IR = prog_instr;
assign prog_cbit = program_cbit;

assign rx_IR = FSL_S_Data;
assign rx_cbit = FSL_S_Control;
assign rx_IR_full = FSL_S_Exists;

// Reads the prog_IR into the ex_IR depending on the mux select:
assign reading_prog_IR =  (((ex_instr_read == 1'b1) 
		      //||((fsl_write == 1'b1) && (mux_select_fifo_in == 2'b01))
			|| ((write_a0 == 1'b1) && (opcode[0] == 1'b1) &&
				(opcode[2] == 1'b0)) || (rx_addr_req == 1'b1))
			&& (sel_ex_IR_input == 1'b0) && (prog_IR_full == 1'b1));

////Used to Read the Program Instructions:
assign prog_instr_read = reading_prog_IR;

// Reads the rx_IR into the ex_IR depending on the mux select:
assign reading_rx_IR = (((ex_instr_read == 1'b1) // unnecessary because this
	//is stored in the base_addr so that the memory can be requested ||
	//		((fsl_write == 1'b1) && (mux_select_fifo_in == 2'b01))
			|| ((write_a0 == 1'b1) && (opcode[0] == 1'b1) &&
				(opcode[2] == 1'b0)) || (rx_addr_req == 1'b1)) 
			&& (sel_ex_IR_input == 1'b1) && (rx_IR_full == 1'b1));

//Used to Read the Slave fsl data (the rx_IR):
//assign FSL_S_Read = reading_rx_IR; // (~rx_IR_full) && FSL_S_Exists;

assign a0_mux_output = ((opcode[0] == 1'b1) && (opcode[2] == 1'b0)) ? 
			ex_IR : new_a0_total;

//Generates the overall controller reset signal and the video reset signal:
assign control_reset = rst || instr_reset;

always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
	instr_reset <= 1'b0;
  else if ((pres_state == execute_instr) && (opcode == 8'h00) && 
	  (exec_prog == 1'b0)) 
	instr_reset <= 1'b1;
  else
	instr_reset <= 1'b0;
end

always @ (opcode or a0_reg or a0_operand) begin
  if (opcode[4] == 1'b1)
    new_a0_total = a0_reg + a0_operand;//ex_IR[0:23];
  else
    new_a0_total = a0_reg - a0_operand;//ex_IR[0:23];
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    a0_reg <= 32'h00000000;
  else if (write_a0 == 1'b1)
    a0_reg <= a0_mux_output;
  else
    a0_reg <= a0_reg;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    a0_operand <= 32'h00000000;
  //else if (change_a0 == 1'b1)
  else if ((ex_instr_read == 1'b1) && (pres_state == get_next_instr))
    a0_operand <= ex_IR[0:23];
  //else if ((ex_IR_full == 1'b1) && (pres_state == get_next_instr))
    //a0_operand <= ex_IR[0:23];
  else
    a0_operand <= a0_operand;
end

always @ (control_reset or change_a0 or Addr_req_W) begin
  if (control_reset == 1'b1)
     write_a0 = 1'b0;
  else if ( (change_a0 == 1'b1) && //(two_part_instr == 1'b0) && 
             (Addr_req_W == 1'b0) )
     write_a0 = 1'b1;
  else 
     write_a0 = 1'b0;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
     change_a0 <= 1'b0;
  else if (write_a0 == 1'b1)
     change_a0 <= 1'b0;
  else if ((pres_state == decode_instr) && ((opcode[3:0] == 4'b1000) ||
	  (opcode[4:1] == 4'b1111) || (opcode[4:0] == 5'b01011)))
     change_a0 <= 1'b1;
  else
     change_a0 <= change_a0;
end

//mem_addr loaded for each new instruction
//incremented for each read/write
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
     mem_addr <= 19'h00000;
  //else if ((enable == 1'b1) || (running == 1'b1))
  else if ((running == 1'b1))
     mem_addr <= mem_addr + 1;
     //mem_addr <= mem_addr + 4;
  else if ((rx_addr_req == 1'b1) && (reading_rx_IR == 1'b1))
     mem_addr <= FSL_S_Data[13:31];
  //else if ((IR_req_W0 == 1'b0) && (IR_req_W1 == 1'b1) && 
  else if ((rx_addr_req == 1'b1) && 
	  (prog_instr_read ==1'b1))
     mem_addr <= prog_instr[13:31];
  else if ((opcode[3] == 1'b1) && (Addr_req_W == 1'b1))
     mem_addr <= a0_reg[13:31];
  //else if ((read_fifo == 1'b1) || (write_fifo == 1'b1))
  else
     mem_addr <= mem_addr;
//READ OR WRITE CYCLE- increment the address!
end

//Need base addr for case when we are writing out data onto the FSL (so 
//we use the correct immediate address):
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    base_addr <= 19'h00000;
  else if ((rx_addr_req == 1'b1) && (reading_rx_IR == 1'b1))
     base_addr <= FSL_S_Data[13:31];
  //else if ((IR_req_W0 == 1'b0) && (IR_req_W1 == 1'b1) && 
  else if ((rx_addr_req == 1'b1) && 
	  (prog_instr_read ==1'b1))
     base_addr <= prog_instr[13:31];
  else 
     base_addr <= base_addr;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    valid_msbs <= 1'b0;
  else if ((rx_addr_req == 1'b1) && 
	  ((reading_rx_IR == 1'b1) || (prog_instr_read == 1'b1)) ||
          ((pres_state == decode_instr) && (opcode[3] == 1'b1)))
     valid_msbs <= 1'b1;
  else if (pres_state == get_next_instr)
     valid_msbs <= 1'b0;
  else 
     valid_msbs <= valid_msbs;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    addr_msb_req <= 13'h0000;
  else if ((rx_addr_req == 1'b1) && (reading_rx_IR == 1'b1))
     addr_msb_req <= FSL_S_Data[0:12];
  else if ((rx_addr_req == 1'b1) && 
	  (prog_instr_read ==1'b1))
     addr_msb_req <= prog_instr[0:12];
  else if ((pres_state == decode_instr) && (opcode[3] == 1'b1))
     addr_msb_req <= a0_reg[0:12];
  else 
     addr_msb_req <= addr_msb_req;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    req_mem <= 1'b0;
  else if ((pres_state == decode_instr) && (opcode[2] == 1'b1) 
	   && ((opcode[1] == 1'b1) || (opcode[0] == 1'b1)) &&
	   ~((exec_prog == 1'b1) && (opcode[1:0] == 2'b10)))
    req_mem <= 1'b1;
  else if ((control_mem0 == 1'b1) || (control_mem1 == 1'b1))
    req_mem <= 1'b0;
  else
    req_mem <= req_mem;
end


always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    IR_req_W0 <= 1'b0;  
//Set IR_req_W0 when you are in the decode stage and the instruction should
//be sent (examples of NOT sent instructions include: wait_rx, write_a0,
//add_a0, sub_a0
  else if ((pres_state == decode_instr) && 
	  (((exec_prog == 1'b1) && (opcode[3:0] != 4'b1100) &&
	  (opcode[3:0] != 4'b1000) && (opcode[3:0] != 4'b1011)) ||
	  (opcode[1:0] == 2'b10)))
    IR_req_W0 <= 1'b1;  
  else if ((fsl_write == 1'b1) && (mux_select_fifo_in == 2'b00)) 
    IR_req_W0 <= 1'b0;  
  else
    IR_req_W0 <= IR_req_W0;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    IR_req_W1 <= 1'b0;  
  else if ((pres_state == decode_instr) && (((exec_prog == 1'b1)
	   && (opcode == `write_Abs_addr)) || (opcode == `readR_Abs_addr)))
    IR_req_W1 <= 1'b1;
  else if ((fsl_write == 1'b1) && (mux_select_fifo_in == 2'b01)) 
	  //&& (IR_req_W0 == 1'b0))
    IR_req_W1 <= 1'b0;
  else
    IR_req_W1 <= IR_req_W1;
end

  
//Instruction includes an address ***FOR NOW DITCH IT***FIX LATER???***
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    rx_addr_req <= 1'b0;  
  else if ((pres_state == decode_instr) && (opcode[4:2] == 3'b001) &&
	  	((opcode[1] == 1'b1) || (opcode[0] == 1'b1)))
	       	//&& (exec_prog == 1'b0))
    rx_addr_req <= 1'b1;
  else if ((rx_addr_req == 1'b1) && 
	  ((reading_rx_IR == 1'b1) || (prog_instr_read == 1'b1)))
    rx_addr_req <= 1'b0;  
  else
    rx_addr_req <= rx_addr_req;
end

//Controls the request to write an address register to the fifo
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    Addr_req_W <= 1'b0;  
  else if ((pres_state == decode_instr) && (opcode[3:2] == 2'b11) &&
  		(opcode[1:0] != 2'b00))
    Addr_req_W <= 1'b1;
  else if ((fsl_write == 1'b1) && (mux_select_fifo_in == 2'b10))
    Addr_req_W <= 1'b0;  
  else
    Addr_req_W <= Addr_req_W;
end

//Should ensure that data is properly read from the fsl to the fifo
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
     rx_data_req <= 1'b0;
  else if ((pres_state == decode_instr) &&  //for sequential execution
		((exec_prog == 1'b0) && (opcode[0] == 1'b1))) 
     rx_data_req <= 1'b1;  
  else if ((data_cntr == 24'h000000) && (done == 1'b1)) 
	  //&& (mux_select_fifo_in == 2'b11))
     rx_data_req <= 1'b0;
  else
     rx_data_req <= rx_data_req;
end

//NOT APPLICABLE FOR THIS APPLICATION- no data to write!
//Controls the request to write data to the fifo
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    data_req_W <= 1'b0;  
  else if ((pres_state == decode_instr) &&  //for sequential execution
		(((exec_prog == 1'b1) && (opcode[1:0] == 2'b11)) || 
	       ((exec_prog == 1'b0) && (opcode[1:0] == 2'b10))) && 
	       (opcode[3:2] != 2'b10)) // && (data_req_W == 1'b0)) 
    data_req_W <= 1'b1;  
  else if ((data_cntr == 24'h000000) && (done == 1'b1)) 
    data_req_W <= 1'b0;  
  else
    data_req_W <= data_req_W;
end
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    new_mem_op <= 1'b0;
  else if (pres_state == get_memory)
    new_mem_op <= 1'b1;
  //else if (enable == 1'b0)
  //else if ((data_cntr == 24'h000001) && (enable == 1'b1))
  else if ((data_cntr == 24'h000001) && (running == 1'b1))
    new_mem_op <= 1'b0;
  else
    new_mem_op <= new_mem_op;
end
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    enable <= 1'b0;
  //else if ((IR_req_W0, IR_req_W1, Addr_req, rx_addr_req) data_req_W, rx_data_req
  //else if ((data_cntr == 24'h000001) && (enable == 1'b1))
  else if ((data_cntr == 24'h000001) && (running == 1'b1))
    enable <= 1'b0;
  //else if ((mux_select_fifo_in == 2'b11)  && (new_mem_op == 1'b1) 
  else if ((pres_state == execute_instr)  && (new_mem_op == 1'b1) 
	    && ((data_req_W == 1'b1) || (rx_data_req == 1'b1)))
    enable <= 1'b1;
  else 
    enable <= enable;
end

//Combine this with the above code for the complete controller
//which may transmit or receive data (eg a memory)
//Loads the number of data values into the counter
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    data_cntr <= 24'h000000;  
  /*else if ((pres_state == decode_instr) &&  //for sequential execution
		(((exec_prog == 1'b1) && (opcode[1:0] == 2'b11)) || 
	       ((exec_prog == 1'b0) && (opcode[1:0] == 2'b10))) && 
	       (opcode[3:2] != 2'b10) && (data_req_W == 1'b0)) */
  else if ((pres_state == decode_instr))  //&&  //for sequential execution
	//	((exec_prog == 1'b0) && (opcode[0] == 1'b1)))  //unnecessary?
    data_cntr <= a0_operand; //ex_IR[0:23];
  //else if ((mux_select_fifo_in == 2'b11) && (fsl_write == 1'b1)) //data out
  //else if ((running == 1'b1) && (enable == 1'b1)) 
  else if ((running == 1'b1)) 
    data_cntr <= data_cntr - 1;
  else
    data_cntr <= data_cntr;
end


// Generate the select signal to the mux feeding the internal buffer:
always @ (pres_state or IR_req_W0 or IR_req_W1 or Addr_req_W or data_req_W) begin
	if ((pres_state == execute_instr) && (IR_req_W0 == 1'b1)) // || 
		//(IR_req_W1 == 1'b1))) 
	      mux_select_fifo_in = 2'b00;
	else if ((pres_state == execute_instr) && (IR_req_W1 == 1'b1)) // && 
		//(IR_req_W1 == 1'b0) && (Addr_req_W == 1'b1))
	      mux_select_fifo_in = 2'b01;
        //else if ( (((pres_state == execute_instr) && (IR_req_W0 == 1'b0) && 
		   //(IR_req_W1 == 1'b0) && (Addr_req_W == 1'b0)) || 
		   //(pres_state == get_next_instr) || 
		   //(pres_state == decode_instr)) && (data_req_W == 1'b1))
	      //mux_select_fifo_in <= 2'b10;
	else if ((pres_state == execute_instr) && (Addr_req_W == 1'b1)) // && 
		//(IR_req_W1 == 1'b0) && (IR_req_W0 == 1'b0))
	      mux_select_fifo_in = 2'b10;
	else 
	      mux_select_fifo_in = 2'b11;
end

// Used to update ex_IR with the next instruction to be executed:
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) //|| (reading_rx_IR == 1'b1))
    //begin
      //ex_IR <= 32'b0;
      ex_cbit <= 1'b0;
    //end
  else if (ex_instr_read == 1'b1) 
    //begin
      //ex_IR <= ex_IR_input;
      ex_cbit <= ex_cbit_input;
    //end
  else
    //begin
      //ex_IR <= ex_IR;
      ex_cbit <= ex_cbit;
    //end
end

// Used to update opcode with the next instruction to be executed:
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    opcode <= 8'b0;
  else if ((ex_instr_read == 1'b1) && (pres_state == get_next_instr))
    opcode <= ex_IR[24:31];
  //else if ((ex_IR_full == 1'b1) && (pres_state == get_next_instr))
    //opcode <= ex_IR[24:31];
  else
    opcode <= opcode;
end
// Used to update ex_IR with the next instruction to be executed:
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) //|| 
    exec_prog <= 1'b1;
  //else if ((two_part_instr == 1'b1) && (ex_instr_read == 1'b1))
    //exec_prog <= exec_prog;
  else if ((ex_instr_read == 1'b1) && (rx_IR_full == 1'b1))
    exec_prog <= 1'b0;
  else if ((ex_instr_read == 1'b1) && (rx_IR_full == 1'b0))
    exec_prog <= 1'b1;
  else
    exec_prog <= exec_prog;
end

// Mux the input into the internal buffer:
always @ (mux_select_fifo_in or ex_IR_fifo_in or a0_reg or mem_out_data or a0_operand or data_req_W or rx_data_req or addr_msb_req or base_addr or addr_msb_ctrl) begin
  case (mux_select_fifo_in) 
    2'b00:
      FSL_out_mux = {a0_operand,ex_IR_fifo_in}; // ex_IR
    2'b01:
      begin
        if ((data_req_W == 1'b0) && (rx_data_req == 1'b0))
          FSL_out_mux = {addr_msb_req,base_addr}; // Imm Addr
        else
          FSL_out_mux = {addr_msb_ctrl,base_addr}; // Imm Addr
      end
    2'b10:
      FSL_out_mux = {addr_msb_ctrl,a0_reg[13:31]}; //a0
    2'b11:
      FSL_out_mux = mem_out_data; //data
    //2'b11:
      //FSL_out_mux = FSL_out_mux;
  endcase
end


// FSM register
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
  //if (reset == 1'b1)
    pres_state <= get_next_instr;
  else
    pres_state <= next_state;
end

// Determine next state
always @ (pres_state or control_reset or exec_prog or opcode or data_req_W or fifo_empty or data_cntr or ex_instr_read or rx_data_req or IR_req_W0 or IR_req_W1 or Addr_req_W or change_a0 or rx_IR_full or control_mem0 or control_mem1 or req_mem or rx_addr_req) begin
  case (pres_state) 
    get_next_instr:
      begin 
	  // Don't attempt to write out this line unless we know for sure
	  // we have space to buffer it. i.e. the svga-line-sized buffer is 
	  // empty
        //if (Vmid == 1'b0 && Hmid_prev[0] == 1'b1 && Hmid == 1'b0 && fifo_exists == 1'b0)  
	if ((ex_instr_read == 1'b1)) // || (ex_IR_full == 1'b1))
	         next_state = decode_instr;
        else
	         next_state = get_next_instr;
      end
    decode_instr:
     /* begin  //Case 1 instr writes out data but the prev instr is using fifo
	if ( ( ((exec_prog == 1'b1) && ((opcode[1:0] == 2'b11) ||
		(opcode[1:0] == 2'b00))) || 
	       ((exec_prog == 1'b0) && (opcode[1:0] == 2'b10)) )
       	    && (opcode[3:2] != 2'b10) && (data_req_W == 1'b1) )
        	  next_state = decode_instr;
	    //Case 2 instr reads in data but the prev instr is using the fsl
	else if ( ((exec_prog == 1'b0) && (opcode[0] == 1'b1))  
       	    && (opcode[3:2] != 2'b10) && (rx_data_req == 1'b1) )
        	  next_state = decode_instr;
	   //There are no shared resource issues so execute
	else*/
		  next_state = get_memory;
      //end
    get_memory:
      begin
	if ((control_mem0 == 1'b1) || (control_mem1 == 1'b1) || 
		(req_mem == 1'b0))
		//((data_req_W == 1'b0) && (rx_data_req == 1'b0)))
	    	  next_state = execute_instr;
	else
	    	  next_state = get_memory;
      end
    execute_instr:
      begin
	  //Continue executing until the instruction/address has been
	  //completed- that data can be written on its own.
	if ((IR_req_W0 == 1'b1) || (IR_req_W1 == 1'b1) || 
	   (Addr_req_W == 1'b1) || (change_a0 == 1'b1) || 
	   //added to remove parallelism
	   (data_req_W == 1'b1) || (rx_data_req == 1'b1)  ||
	   (rx_addr_req == 1'b1) || 
	   //((exec_prog == 1'b0) && (opcode[0] == 1'b1) &&
	   //(fifo_empty == 1'b0)) ||  //just fifo empty???
	   ((opcode == `wait_rx) && (rx_IR_full == 1'b0)))
		  next_state = execute_instr;
	else
		  next_state = get_next_instr;
        end
  endcase
end

always @ (mux_select_fifo_in or IR_req_W0 or IR_req_W1 or Addr_req_W or data_req_W or fifo_full or write_fifo or FSL_M_Full or pres_state or exec_prog or ex_IR_full or rx_addr_req) begin
  case(mux_select_fifo_in)
     //*****Do not worry about back to back writes******************
     2'b00:
       begin
         if ((IR_req_W0 == 1'b1) && (FSL_M_Full == 1'b0) &&
	 	(pres_state == execute_instr)) 
		 // && ((data_req_W == 1'b0) ||
		//((data_req_W == 1'b1) && (Hmid == 1'b0)))) 
		 //(fifo_exists == 1'b0)) // &&
		 //(fsl_write == 1'b0))
		 // (pres_state == execute_instr) &&
		// || (IR_req_W1 == 1'b1)) && (ex_IR_full == 1'b1) && 
		 //(fifo_full == 1'b0)) 
            fsl_write = 1'b1;
	 else 
            fsl_write = 1'b0;
       end
     2'b01:
       begin
         if ((IR_req_W1 == 1'b1) && 
		 (((exec_prog == 1'b1) && (ex_IR_full == 1'b1)) ||
		  ((exec_prog == 1'b0) && (rx_addr_req == 1'b0))) &&
		 (FSL_M_Full == 1'b0) && (pres_state == execute_instr)) 
		 // && (fsl_write == 1'b0))
            fsl_write = 1'b1;
	 else 
            fsl_write = 1'b0;
       end
     2'b10:
       begin
         if ((Addr_req_W == 1'b1) && (FSL_M_Full == 1'b0)
		 && (pres_state == execute_instr)) // && 
		 //(fsl_write == 1'b0))
            fsl_write = 1'b1;
	 else 
            fsl_write = 1'b0;
       end
     2'b11:
       /*begin
         if ((data_req_W == 1'b1) && (fifo_full == 1'b0) && 
		 (valid_data == 1'b1)) // && (fsl_write == 1'b0))
            fsl_write = 1'b1;
	 else */
            //fsl_write = 1'b0;
       begin
	 if (pres_state == execute_instr) /* shouldn't need full/empty as they  
		 			are part of the mem controller */
            fsl_write = write_fifo;
	 else
	    fsl_write = 1'b0;
       end
  endcase
end


//// FSM register
//always @ (posedge pixel_clock or posedge reset) begin
  //if (reset == 1'b1) begin
    //pres_state <= st_reset; 
  //end
  //else begin
    //pres_state <= next_state;
  //end
//end
//
//// Determine next state
//always @ (negedge pixel_clock or posedge reset) begin
  //if (reset == 1'b1) begin
    //next_state <= st_reset; 
  //end
  //else begin
    //case (pres_state) 
      //st_reset:
	//begin
          //if (reset == 1'b0) begin
            //next_state <= st_ctrl;
	  //end
	  //else begin
            //next_state <= st_reset;
	  //end
	//end
      //st_ctrl:
	//next_state <= st_addr;
      //st_addr:
	//next_state <= st_read;
      //st_read:
	//begin
          //if (user_access_ok == 1'b1) begin  // Keep reading until all data is back
            //next_state <= st_useracc_ok;
	  //end
	  //else begin
            //next_state <= st_read;
	  //end
	//end
      //st_useracc_ok:
	//begin
          //if (user_access_ok == 1'b1) begin  
            //next_state <= st_useracc_ok;
	  //end
	  //else begin
            //next_state <= st_ctrl;
	  //end
	//end
    //endcase
  //end
//end
//

/*
fsl_mem0_program fsl_mem0_program (
  // inputs
  .clk(clk), 
  .rst(control_reset), 
  
  //Program Interface:
  .prog_instr(prog_instr),
  .program_cbit(program_cbit),
  .prog_instr_read(prog_instr_read),
  .valid_instr(prog_IR_full),
  .sbit_using_mem0(sbit_using_mem0)  //Status bit 0- controller has access to
  					//mem bank 0
  );
*/


endmodule //SIMPPL_CONTROLLER_FOR_A_MEMORY_BANK
