//************************************************************************
//    This file was written for University of Toronto, ECE532 
//    File: fsl_video_input.v
//    Uses: coregen_fifo.xco (and .edn, but .xco goes in the project & ISE
//            will look for .edn during compilation)
//          addr_gen.v
//          input2rgb.v
//    Used by: top_dualzbt_dow.v
//    Works with ProjectNavigator 6.2.03i
//    Description: Wraps video input processing modules and provides an FSL
//    interface.  Requires the line-length FIFO to avoid losing data and
//    clogging up the FSL.  
//
//    Protocol for communication is:
//      control word:  upper 19 bits give # of words to write out
//                     20th bit is read/write_n
//      address word:  upper 19 bits give address (19 bits to address into
//                      a ZBT)
//      data:  must output exactly the number of data words you specified
//              you would in the control word, otherwise the FSLs get clogged
//
//    NCKC August 2004
//************************************************************************ 
//
 

module fsl_video_input (
  // inputs
  clk, rst, // rst should be driven by FSL_Rst
  VIDEO_INPUT,
  //Master FSL
  FSL_M_Full,
  FSL_M_Clk,
  FSL_M_Data,
  FSL_M_Control,
  FSL_M_Write,
  //Slave FSL
  FSL_S_Clk,
  FSL_S_Read,
  FSL_S_Data,
  FSL_S_Control,
  FSL_S_Exists,
  // outputs
  release_priority,
  led0_o, led1_o, // for debug
  batch
  ////FOR TESTING ONLY:
  //Vmid,
  //Hmid,
  //Fmid
  );

  input clk, rst;
  input [7:0] VIDEO_INPUT; // Connect the MSB pins
  input FSL_M_Full;
  output FSL_M_Clk, FSL_M_Control, FSL_M_Write;
//  output [31:0] FSL_M_Data;
  output [0:31] FSL_M_Data;
  input FSL_S_Control, FSL_S_Exists;
  input [0:31] FSL_S_Data;
  output FSL_S_Clk, FSL_S_Read;
  output release_priority; // Could be used to demand priority servicing. 1 => give up priority, 0 => demand priority
  output batch;  // 1 => Still on the same frame
  output led0_o, led1_o; // for debug
  //FOR TESTING ONLY:
  //input Vmid, Hmid, Fmid;
	  
  wire clk, rst;
  wire [7:0] VIDEO_INPUT;
  wire [7:0] R,G,B;
  //wire release_priority;  // Do not use
  wire FSL_M_Clk, FSL_M_Full;
  wire FSL_M_Control;
  reg FSL_M_Write, FSL_M_Writed;
  wire [0:31] FSL_M_Data;
//  wire [31:0] FSL_M_Data;
  
  //Slave FSL wires:
  wire FSL_S_Clk;
  wire FSL_S_Read;
  wire [0:31] FSL_S_Data;
  wire FSL_S_Control;
  wire FSL_S_Exists;

  //IR signals:
  //reg  rx_IR_full;
  //reg [0:31] rx_IR;
  //reg  rx_cbit;  //receive control bit
  wire  rx_IR_full;
  wire [0:31] rx_IR;
  wire  rx_cbit;  //receive control bit
  //reg prog_IR_full;
  //reg [0:31] prog_IR;
  //reg prog_cbit;  //program control bit
  wire prog_IR_full;
  wire [0:31] prog_IR;
  wire prog_cbit;  //program control bit
  //reg  ex_IR_full;
  //reg [0:31] ex_IR;
  reg  ex_cbit;  //executing control bit
  reg exec_prog; //am I executing the program or an rx_IR**only valid when 
  			//ex_IR_full is high

  //Address registers:
  reg [0:31] a0_reg; //Address register 0 
  reg [0:23] a0_operand; //latches the data portion of the instruction
  reg write_a0;

  //Mux control signals:
  wire [0:31] a0_mux_output; //multiplexer output to the a0 reg
  reg [0:31] fifo_mux_out; //output from multiplexer into buffering fifo
  //wire IR_sel_exec_path_mux; // Used to determine whether the next value read
  				//into the ex_IR is the second part of an
				//instruction
 
  wire control_reset;  //combines system reset and instruction reset
  //wire instr_reset;	//generated via the reset instruction- asynch pulse
  reg instr_reset;	//generated via the reset instruction- clock period long

  //Used to interface the ex_IR, A0 and data with the fifo: 
  reg [1:0] mux_select_fifo_in; //used to select
  //reg [0:7] ex_IR_fifo_in;
  wire [0:7] ex_IR_fifo_in;
  //reg decoded_IR; //Used to indicate that a valid instruction has been decoded
  reg [7:0] opcode;  //Used to store the opcode for 2-part instructions
  reg valid_data_enable;  //Used to ensure only one line gets written to memory
  				//between H_synchs
  reg valid_data_reg; //Used to indicate when Hmid signal indicates valid data
  wire valid_data;  //combines the Hmid info with the video state info

  //Control registers:	
  reg change_a0;  //Flag to indicate when instruction will alter the value of a0
  //reg two_part_instr; // Indicates that there is a 2nd part to the instruction
  reg IR_req_W0;  // Indicates when the instruction writes to the fifo
  reg IR_req_W1;  // Indicates when the instruction writes an immed to the fifo
  reg Addr_req_W;  // Indicates when the address register needs to be written to
  				//the fifo
  reg data_req_W;  // Indicates that there is data to be written to the fifo
  reg rx_data_req; // Used to indicate when data should be read in as part of
  			// the instruction.
  reg [0:31] new_a0_total;  //Output from Accumulator;

  //Program interface and Instruction Register Read signals:
  wire [0:31] prog_instr;  //connects to the program instrucion input port
  wire program_cbit; //connects to the program's control bit
  wire prog_instr_read; //reading from the port into the program IR
  wire done_field; //status bit provided to the program for branching
  wire done_frame; //status bit provided to the program for branching
  wire video_in_on; //status bit provided to program for initialization
  //wire [0:31] ex_IR_input;  //wire used as mux output from prog and rx IRs
  wire [0:31] ex_IR;  //wire used as mux output from prog and rx IRs
  wire ex_cbit_input; //wire used as mux output from prog and rx cbits
  wire ex_instr_read;  //reads an instruction into the ex_IR
  wire reading_prog_IR;  //reading the prog_IR clears it
  wire reading_rx_IR;  //reading the rx_IR clears it
  //wire reading_ex_IR; //reading the ex_IR clears it
  wire sel_ex_IR_input; // selects the input to the ex_IR
  wire rx_req_W;

  reg batch;
//
  //reg FSL_M_Controlval; //FSL_M_Writeval;
//  reg [31:0] FSL_M_Dataval;
//
  wire Fmid, Vmid, Hmid;
  //wire [18:0] Addr;
  //reg [18:0] Addr_prev;
  wire [31:0] RGBdata;
//  wire [23:0] RGBdata;
//
  reg [1:0] pres_state;
  reg [1:0] next_state;
  reg [2:0] pres_v_state;
  reg [2:0] next_v_state;
  reg [5:0] Hmid_prev;
  reg Vmid_prev;

// Used to store status of the rx_IR:
  // For internal FIFO
  reg fifo_read, fifo_write; 
  wire fifo_exists, fifo_full;
  //wire [24:0] DataIn, DatatoFSL; // Is this correct endianness?
  wire [32:0] DatatoFSL, DataIn; // Is this correct endianness?
  reg wrote_last;  // == 1 if wrote in last cycle, otherwise didn't write 
  //
  reg [23:0] pixel_ct; // for debugging
//
  //reg led0_o, led1_o; // for debug  
  wire led0_o, led1_o; // for debug  
////  reg [31:0] RGBdata_reg;
//
//// For Coregen FIFO
wire fifo_empty;
assign fifo_exists = ~fifo_empty;

//assign rd_prog_instr = 1'b0;
//assign done_field = 1'b0;
//assign done_frame = 1'b0;


assign ex_IR_fifo_in[0:4] = 5'b00000;
assign ex_IR_fifo_in[5:7] = opcode[2:0];

assign done_field = Vmid;  //When vertical lines are blanking, in between fields (valid data)
assign done_frame = Fmid && Vmid && ~Vmid_prev; //Just entered blanking from even field
assign video_in_on = pres_v_state[2] || pres_v_state[1];  //as long as we're not
			// in reset or wait state, video is up and // running

//always @ (Fmid or Vmid or Vmid_prev) begin
  //if (Fmid && Vmid && ~Vmid_prev)
    //done_frame <= 1'b1;
  //if (Fmid && Vmid)
    //done_frame <= done_frame;
  //else 
    //done_frame <= 1'b0;
//end

// NOTE: Changing the number of states will probably require changing how 
// Hmid_prev is used
//parameter  st_reset=20;  
//parameter  st_ctrl=21;
//parameter  st_ctrl2=22;
//parameter  st_addr=23;
//parameter  st_addr2=24;
//parameter  st_data=25;
//parameter  st_data2=26; // Because pixels come in at 1/2 clk rate (422->444)

//`define control_word 32'h00028005 //Used by the SIMPPL protocol
////`define control_word 32'h00500000 //Used in Nathalie's protocol
`define line_length 10'd640 

//States for overall execution path state machine:
parameter get_next_instr=0;
parameter decode_instr=1;
parameter execute_instr=2;
//parameter wait_rx_instr=3;

//States for video valid data signal state machine:
parameter st_reset=0;
parameter st_wait=1;
parameter st_blankA=2;
parameter st_blankB=3;
parameter st_odd=4;
parameter st_even=5;

//Create constants for all the instruction opcode:
`define reset 8'h00
`define noop 8'h04

`define readR_Imm 8'h02
`define write_Imm 8'h03
`define readA_Imm 8'h01

`define readR_Abs_addr 8'h06  
`define write_Abs_addr 8'h07  // These first 8 instructions will be 
`define readA_Abs_addr 8'h05  // read and transmitted.

`define read_Areg 8'h0A
`define write_Areg 8'h0B

`define add_imm_Areg 8'h18
`define sub_imm_Areg 8'h08
`define wait_rx 8'h0C

`define read_Areg_indirect 8'h0E
`define write_Areg_indirect 8'h0F

`define read_Areg_autoinc 8'h1E
`define write_Areg_autoinc 8'h1F

assign release_priority = 8'b1;

assign rx_req_W = 1'b0;

assign FSL_M_Data = DatatoFSL[31:0];
assign FSL_M_Control = DatatoFSL[32];
assign DataIn[32] = (mux_select_fifo_in == 2'b00) ? IR_req_W0 : 1'b0; // FSL_M_Controlval;
//assign DataIn[32] = (mux_select_fifo_in == 2'b00) ? ex_cbit : ex_cbit_input; // FSL_M_Controlval;

assign valid_data = ((valid_data_reg == 1'b1) && (data_req_W == 1'b1));
			// && (Hmid_prev[4] == 1'b0)); && //necessary?
		//((pres_v_state == st_even) || (pres_v_state == st_odd)));
assign DataIn[31:0] = fifo_mux_out;  //FSL_M_Dataval;
  
//Used to Read an instruction into the Executing IR:
//assign ex_instr_read = (rx_IR_full || prog_IR_full) && (~ex_IR_full);
assign ex_instr_read = ((pres_state == get_next_instr)
	 && ((rx_IR_full || prog_IR_full))); 
		// && (fifo_exists == 1'b0));

//Used to mux the outputs from the two different IRs:
//assign sel_ex_IR_input = two_part_instr ? ~exec_prog : rx_IR_full;
assign sel_ex_IR_input = (pres_state == get_next_instr) ? 
				rx_IR_full : ~exec_prog;
//assign ex_IR_input = sel_ex_IR_input ? rx_IR : prog_IR ;
assign ex_IR = sel_ex_IR_input ? rx_IR : prog_IR ;
assign ex_cbit_input = sel_ex_IR_input ? rx_cbit : prog_cbit ;
   
assign prog_IR = prog_instr;
assign prog_cbit = program_cbit;

assign rx_IR = FSL_S_Data;
assign rx_cbit = FSL_S_Control;
assign rx_IR_full = FSL_S_Exists;
   
// Reads the prog_IR into the ex_IR depending on the mux select:
//assign reading_prog_IR = ex_instr_read && ~sel_ex_IR_input;
assign reading_prog_IR =  (((ex_instr_read == 1'b1) ||
			((fifo_write == 1'b1) && (mux_select_fifo_in == 2'b01))
			|| ((write_a0 == 1'b1) && (opcode[0] == 1'b1) &&
				(opcode[2] == 1'b0)))
			&& (sel_ex_IR_input == 1'b0) && (prog_IR_full == 1'b1));

//Used to Read the Program Instructions:
//assign prog_instr_read = ~prog_IR_full;
assign prog_instr_read = reading_prog_IR;

// Reads the rx_IR into the ex_IR depending on the mux select:
//assign reading_rx_IR = ex_instr_read && sel_ex_IR_input;
//assign reading_rx_IR = ex_instr_read && sel_ex_IR_input;
assign reading_rx_IR = (((ex_instr_read == 1'b1) ||
			((fifo_write == 1'b1) && (mux_select_fifo_in == 2'b01))
			|| ((write_a0 == 1'b1) && (opcode[0] == 1'b1) &&
				(opcode[2] == 1'b0)))
			&& (sel_ex_IR_input == 1'b1) && (rx_IR_full == 1'b1));

//Used to Read the Slave fsl data:
assign FSL_S_Read = reading_rx_IR; // (~rx_IR_full) && FSL_S_Exists;
  
// Reads the ex_IR when writing to the fifo from position 0:
//assign reading_ex_IR = fifo_write && (mux_select_fifo_in == 2'b00); 

assign FSL_M_Clk = clk;
assign FSL_S_Clk = clk;
assign RGBdata[31:24] = 8'b0;
assign R = RGBdata[23:16];  
assign G = RGBdata[15:8];  
assign B = RGBdata[7:0];  

//assign opcode = ex_IR[24:31];

assign a0_mux_output = ((opcode[0] == 1'b1) && (opcode[2] == 1'b0)) ? 
			ex_IR : new_a0_total;

//Generates the overall controller reset signal:
assign control_reset = rst || instr_reset;

//Used to generate the controller's reset signal from the reset instruction:
//assign instr_reset = ((pres_state == execute_instr) && (opcode == 8'h00) &&
	  		//(exec_prog == 1'b0)) ? 1'b1 : 1'b0;

always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
	instr_reset <= 1'b0;
  else if ((pres_state == execute_instr) && (opcode == 8'h00) && 
	  (exec_prog == 1'b0)) 
	instr_reset <= 1'b1;
  else
	instr_reset <= 1'b0;
end

always @ (opcode or a0_reg or a0_operand) begin
  if (opcode[4] == 1'b1)
    new_a0_total = a0_reg + a0_operand;//ex_IR[0:23];
  else
    new_a0_total = a0_reg - a0_operand;//ex_IR[0:23];
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    a0_reg <= 32'h00000000;
  else if (write_a0 == 1'b1)
    a0_reg <= a0_mux_output;
  else
    a0_reg <= a0_reg;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    a0_operand <= 32'h00000000;
  //else if (change_a0 == 1'b1)
  else if ((ex_instr_read == 1'b1) && (pres_state == get_next_instr))
    a0_operand <= ex_IR[0:23];
  //else if ((ex_IR_full == 1'b1) && (pres_state == get_next_instr))
    //a0_operand <= ex_IR[0:23];
  else
    a0_operand <= a0_operand;
end

always @ (control_reset or change_a0 or Addr_req_W) begin
  if (control_reset == 1'b1)
     write_a0 = 1'b0;
  else if ( (change_a0 == 1'b1) && //(two_part_instr == 1'b0) && 
             (Addr_req_W == 1'b0) )
     write_a0 = 1'b1;
  else 
     write_a0 = 1'b0;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
     change_a0 <= 1'b0;
  else if (write_a0 == 1'b1)
     change_a0 <= 1'b0;
  else if ((pres_state == decode_instr) && ((opcode[3:0] == 4'b1000) ||
	  (opcode[4:1] == 4'b1111) || (opcode[4:0] == 5'b01011)))
     change_a0 <= 1'b1;
  else
     change_a0 <= change_a0;
end

///***************NEEDS MORE FOR RECEIVE***************
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
     rx_data_req <= 1'b0;
  else
     rx_data_req <= rx_data_req;
end
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    IR_req_W0 <= 1'b0;  
//Set IR_req_W0 when you are in the decode stage and the instruction should
//be sent (examples of NOT sent instructions include: wait_rx, write_a0,
//add_a0, sub_a0
  else if ((pres_state == decode_instr) && 
	  (((exec_prog == 1'b1) && (opcode[3:0] != 4'b1100) &&
	  (opcode[3:0] != 4'b1000) && (opcode[3:0] != 4'b1011)) ||
	  (opcode[1:0] == 2'b10)))
    IR_req_W0 <= 1'b1;  
  else if ((fifo_write == 1'b1) && (mux_select_fifo_in == 2'b00)) 
    IR_req_W0 <= 1'b0;  
  else
    IR_req_W0 <= IR_req_W0;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    IR_req_W1 <= 1'b0;  
  else if ((pres_state == decode_instr) && (((exec_prog == 1'b1)
	   && (opcode == `write_Abs_addr)) || (opcode == `readR_Abs_addr)))
    IR_req_W1 <= 1'b1;
  else if ((fifo_write == 1'b1) && (mux_select_fifo_in == 2'b01)) 
	  //&& (IR_req_W0 == 1'b0))
    IR_req_W1 <= 1'b0;
  else
    IR_req_W1 <= IR_req_W1;
end

//Controls the request to write an address register to the fifo
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    Addr_req_W <= 1'b0;  
  else if ((pres_state == decode_instr) && (opcode[3:2] == 2'b11) &&
  		(opcode[1:0] != 2'b00))
    Addr_req_W <= 1'b1;
  else if ((fifo_write == 1'b1) && (mux_select_fifo_in == 2'b10))
    Addr_req_W <= 1'b0;  
  else
    Addr_req_W <= Addr_req_W;
end

//Controls the request to write data to the fifo
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    data_req_W <= 1'b0;  
  //else if ((pres_state == execute_instr) && //for parallel execution
  else if ((pres_state == decode_instr) &&  //for sequential execution
		(((exec_prog == 1'b1) && (opcode[1:0] == 2'b11)) || 
	       ((exec_prog == 1'b0) && (opcode[1:0] == 2'b10))) && 
	       (opcode[3:2] != 2'b10) && (data_req_W == 1'b0)) 
    data_req_W <= 1'b1;  
  else if ((pixel_ct == 24'h000001) && (fifo_write == 1'b1)) 
	  //&& (mux_select_fifo_in == 2'b11))
    data_req_W <= 1'b0;
  else
    data_req_W <= data_req_W;
end

//Loads the number of data values into the counter
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    pixel_ct <= 24'h000000;  
  //else if ((pres_state == execute_instr) && //for parallel execution
  else if ((pres_state == decode_instr) &&  //for sequential execution
		(((exec_prog == 1'b1) && (opcode[1:0] == 2'b11)) || 
	       ((exec_prog == 1'b0) && (opcode[1:0] == 2'b10))) && 
	       (opcode[3:2] != 2'b10) && (data_req_W == 1'b0)) 
    pixel_ct <= a0_operand; //ex_IR[0:23];
  else if ((mux_select_fifo_in == 2'b11) && (fifo_write == 1'b1)) 
    pixel_ct <= pixel_ct - 1;
  else
    pixel_ct <= pixel_ct;
end

// Used to set the register used to determine if the instruction has been 
// decoded.
//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset == 1'b1)
	  //two_part_instr <= 1'b0;
  //else if (ex_instr_read == 1'b1)
	  //two_part_instr <= 1'b0;
  ////else if ((ex_instr_read_prev = 1'b1) && ( ex_instr_read = 1'b1))
  //else if (( ex_IR_full == 1'b1) && (pres_state == decode_instr) &&
	  ////(ex_cbit == 1'b1) &&
	   //((opcode == `readR_Abs_addr) || (opcode == `write_Abs_addr) || 
	   //(opcode == `readA_Abs_addr) || (opcode == `write_Areg)))
	  //two_part_instr <= 1'b1;
  //else
	  //two_part_instr <= two_part_instr;
//end

//// Used to mux the appropriate versions of of the ex_IR:
//always @ (IR_req_W0 or ex_IR) begin
   //if (IR_req_W0 ==1'b1) 
	//begin
	   //ex_IR_fifo_in[0:4] <= 5'b00000;
   	   //ex_IR_fifo_in[5] <= ex_IR[29];
	   //ex_IR_fifo_in[6:7] <= ex_IR[30:31]; //**RX VS TX**
	//end
   //else
	   //ex_IR_fifo_in <= ex_IR[24:31];
//end

// Generate the select signal to the mux feeding the internal buffer:
always @ (pres_state or IR_req_W0 or IR_req_W1 or Addr_req_W or data_req_W) begin
	if ((pres_state == execute_instr) && (IR_req_W0 == 1'b1)) // || 
		//(IR_req_W1 == 1'b1))) 
	      mux_select_fifo_in = 2'b00;
	else if ((pres_state == execute_instr) && (IR_req_W1 == 1'b1)) // && 
		//(IR_req_W1 == 1'b0) && (Addr_req_W == 1'b1))
	      mux_select_fifo_in = 2'b01;
        //else if ( (((pres_state == execute_instr) && (IR_req_W0 == 1'b0) && 
		   //(IR_req_W1 == 1'b0) && (Addr_req_W == 1'b0)) || 
		   //(pres_state == get_next_instr) || 
		   //(pres_state == decode_instr)) && (data_req_W == 1'b1))
	      //mux_select_fifo_in <= 2'b10;
	else if ((pres_state == execute_instr) && (Addr_req_W == 1'b1)) // && 
		//(IR_req_W1 == 1'b0) && (IR_req_W0 == 1'b0))
	      mux_select_fifo_in = 2'b10;
	else 
	      mux_select_fifo_in = 2'b11;
end

//// Used to store status of the rx_IR:
//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset == 1'b1) 
    //prog_IR_full <= 1'b0;
  //else if (reading_prog_IR == 1'b1)
    //prog_IR_full <= 1'b0;
  //else if (prog_instr_read == 1'b1)
    //prog_IR_full <= 1'b1;
  //else
    //prog_IR_full <= prog_IR_full;
//end
//
//// Used to Read Slave FSL data into rx_IR:
//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset == 1'b1) 
    //begin
      //prog_IR <= 32'b0;
      //prog_cbit <= 1'b0;
    //end
  //else if (prog_instr_read == 1'b1) 
    //begin
      //prog_IR <= prog_instr;
      //prog_cbit <= program_cbit;
    //end
  //else
    //begin
      //prog_IR <= prog_IR;
      //prog_cbit <= prog_cbit;
    //end
//end

//// Used to store status of the rx_IR:
//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset == 1'b1) 
    //rx_IR_full <= 1'b0;
  //else if (reading_rx_IR == 1'b1)
    //rx_IR_full <= 1'b0;
  //else if (FSL_S_Read == 1'b1)
    //rx_IR_full <= 1'b1;
  //else
    //rx_IR_full <= rx_IR_full;
//end
//
//// Used to Read Slave FSL data into rx_IR:
//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset == 1'b1)
    //begin
      //rx_IR <= 32'b0;
      //rx_cbit <= 1'b0;
    //end
  //else if (FSL_S_Read == 1'b1) 
    //begin
      //rx_IR <= FSL_S_Data;
      //rx_cbit <= FSL_S_Control;
    //end
  //else
    //begin
      //rx_IR <= rx_IR;
      //rx_cbit <= rx_cbit;
    //end
//end

//// Used to store status of the ex_IR:
//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset == 1'b1) 
    //ex_IR_full <= 1'b0;
////NOT WORKING:
  ////else if ((reading_ex_IR == 1'b1) || ((pres_state == execute_instr) &&
	  ////(IR_req_W0 == 1'b0) && (IR_req_W1 == 1'b0) && (Addr_req_W == 1'b0)
	  ////&& (((change_a0 == 1'b1) && (two_part_instr == 1'b1)) ||
	  ////((change_a0 == 1'b0) && (two_part_instr == 1'b0)))))
  //else if (ex_instr_read == 1'b1)
    //ex_IR_full <= 1'b1;
////WORKING:
  //else if ((reading_ex_IR == 1'b1) || ((pres_state == execute_instr) &&
	  //(opcode == `write_Areg) && (two_part_instr == 1'b1)) ||
	  //((pres_state == execute_instr) && ((opcode == `write_Areg) ||
	  //(opcode == `add_imm_Areg) || (opcode == `sub_imm_Areg)) && 
	  //(change_a0 == 1'b0)) || ((pres_state == execute_instr) && 
	  //(((opcode == `wait_rx) && (rx_IR_full == 1'b1)) || 
  	  //((exec_prog == 1'b0) && (opcode[1:0] == 2'b00))))) 
	  ////add that the ex_instr_read should be full?****
    //ex_IR_full <= 1'b0;
  //else
    //ex_IR_full <= ex_IR_full;
//end

// Used to update ex_IR with the next instruction to be executed:
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) //|| (reading_rx_IR == 1'b1))
    //begin
      //ex_IR <= 32'b0;
      ex_cbit <= 1'b0;
    //end
  else if (ex_instr_read == 1'b1) 
    //begin
      //ex_IR <= ex_IR_input;
      ex_cbit <= ex_cbit_input;
    //end
  else
    //begin
      //ex_IR <= ex_IR;
      ex_cbit <= ex_cbit;
    //end
end

// Used to update opcode with the next instruction to be executed:
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    opcode <= 8'b0;
  else if ((ex_instr_read == 1'b1) && (pres_state == get_next_instr))
    opcode <= ex_IR[24:31];
  //else if ((ex_IR_full == 1'b1) && (pres_state == get_next_instr))
    //opcode <= ex_IR[24:31];
  else
    opcode <= opcode;
end
// Used to update ex_IR with the next instruction to be executed:
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) //|| 
    exec_prog <= 1'b1;
  //else if ((two_part_instr == 1'b1) && (ex_instr_read == 1'b1))
    //exec_prog <= exec_prog;
  else if ((ex_instr_read == 1'b1) && (rx_IR_full == 1'b1))
    exec_prog <= 1'b0;
  else if ((ex_instr_read == 1'b1) && (rx_IR_full == 1'b0))
    exec_prog <= 1'b1;
  else
    exec_prog <= exec_prog;
end

// Mux the input into the internal buffer:
always @ (mux_select_fifo_in or ex_IR or ex_IR_fifo_in or a0_reg or RGBdata or a0_operand) begin
  case (mux_select_fifo_in) 
    2'b00:
      fifo_mux_out = {a0_operand,ex_IR_fifo_in};
    2'b01:
      fifo_mux_out = ex_IR; //{13'h0,Addr};
    2'b10:
      fifo_mux_out = a0_reg;
    2'b11:
      fifo_mux_out = RGBdata;
    //2'b11:
      //fifo_mux_out = fifo_mux_out;
  endcase
end


//
// Use Horizontal blanking signal to control state changes
// Also register RGB values to give delay
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) begin
    Hmid_prev <= 6'b000000;
//    RGBdata_reg <= 32'h0;
  end
  else begin
    Hmid_prev[5:1] <= Hmid_prev[4:0];
    Hmid_prev[0] <= Hmid;
//    RGBdata_reg <= RGBdata;
  end
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
	  Vmid_prev <= 1'b0;
  else 
	  Vmid_prev <= Vmid;
end

//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset  == 1'b1)
    //valid_data_reg <= 1'b0;
  //else if ((Hmid_prev[1] == 1'b1) && (Hmid_prev[0] == 1'b0) && 
	  //((pres_v_state == st_odd) || (pres_v_state == st_even)))
    //valid_data_reg <= 1'b1;
  //else if ((Hmid_prev[1] == 1'b0) && //(Hmid == 1'b0) && 
	  //((pres_v_state == st_odd) || (pres_v_state == st_even)))
    //valid_data_reg <= ~valid_data_reg;
  //else 
    //valid_data_reg <= 1'b0;
//end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset  == 1'b1)
    valid_data_enable <= 1'b0;
  else if ((Hmid_prev[4] == 1'b1) && (Hmid_prev[3] == 1'b0) && 
	  (data_req_W == 1'b1) &&
	  ((pres_v_state == st_odd) || (pres_v_state == st_even)))
    valid_data_enable <= 1'b1;
  else if (data_req_W == 1'b0) 
    valid_data_enable <= 1'b0;
  else 
    valid_data_enable <= valid_data_enable;
end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset  == 1'b1)
    valid_data_reg <= 1'b0;
  //else if ((Hmid_prev[5] == 1'b1) && (valid_data_enable == 1'b1))
    //valid_data_reg <= 1'b1;
  //else if ((Hmid_prev[5] == 1'b0) && (valid_data_enable == 1'b1)) 
    //valid_data_reg <= ~valid_data_reg;
  else if ((Hmid_prev[5] == 1'b1) && (Hmid_prev[4] == 1'b0)
	  && (valid_data_enable == 1'b1))
    valid_data_reg <= 1'b1;
  else if (((Hmid_prev[5] == 1'b0) || (Hmid_prev[4] == 1'b1))
	  && (valid_data_enable == 1'b1)) 
    valid_data_reg <= ~valid_data_reg;
  else 
    valid_data_reg <= 1'b0;
end


// Video input status FSM register
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
  //if (rst == 1'b1)
    pres_v_state <= st_reset;
  else
    pres_v_state <= next_v_state;
end

//Determine next state for Video input FSM:
always @ (pres_v_state or Fmid or Vmid or Hmid or control_reset) begin
  case(pres_v_state)
    st_reset:
      begin
	if ((Fmid == 1'b0) && (Vmid == 1'b1)) //may be entering odd field
		next_v_state = st_wait;
	else
		next_v_state = st_reset;
      end
    st_wait:
      begin
	if ((Fmid == 1'b0) && (Vmid == 1'b0)) //entering odd field
		next_v_state = st_odd;
	else if ((Fmid == 1'b1) && (Vmid == 1'b1)) //haven't finished even field
		next_v_state = st_reset;	//so go back to reset
	else
		next_v_state = st_wait;
      end
    st_odd:
      begin
	if ((Fmid == 1'b0) && (Vmid == 1'b1)) //entering blanking "A" state
		next_v_state = st_blankA;
	else
		next_v_state = st_odd;
      end
    st_blankA:
      begin
	if ((Fmid == 1'b1) && (Vmid == 1'b1)) //entering blanking "B" state
		next_v_state = st_blankB;
	else if ((Fmid == 1'b0) && (Vmid == 1'b0)) //entering odd field
		next_v_state = st_odd;
	else
		next_v_state = st_blankA;
      end
    st_blankB:
      begin
	if ((Fmid == 1'b1) && (Vmid == 1'b0)) //entering even field
		next_v_state = st_even;
	else if ((Fmid == 1'b0) && (Vmid == 1'b1)) //entering blanking "A" state
		next_v_state = st_blankA;
	else
		next_v_state = st_blankB;
      end
    st_even:
      begin
	if ((Fmid == 1'b1) && (Vmid == 1'b1)) //entering blanking "B" state
		next_v_state = st_blankB;
	else
		next_v_state = st_even;
      end
  endcase
end
// FSM register
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
  //if (rst == 1'b1)
    pres_state <= get_next_instr;
  else
    pres_state <= next_state;
end

// Determine next state
always @ (pres_state or control_reset or exec_prog or opcode or data_req_W or fifo_exists or pixel_ct or ex_instr_read or rx_req_W or IR_req_W0 or IR_req_W1 or Addr_req_W or change_a0 or rx_IR_full) begin
  case (pres_state) 
    get_next_instr:
      begin 
	  // Don't attempt to write out this line unless we know for sure
	  // we have space to buffer it. i.e. the svga-line-sized buffer is 
	  // empty
        //if (Vmid == 1'b0 && Hmid_prev[0] == 1'b1 && Hmid == 1'b0 && fifo_exists == 1'b0)  
	if ((ex_instr_read == 1'b1)) // || (ex_IR_full == 1'b1))
	         next_state = decode_instr;
        else
	         next_state = get_next_instr;
      end
    decode_instr:
      begin  //Case 1 instr writes out data but the prev instr is using fifo
	if ( ( ((exec_prog == 1'b1) && ((opcode[1:0] == 2'b11) ||
		(opcode[1:0] == 2'b00))) || 
	       ((exec_prog == 1'b0) && (opcode[1:0] == 2'b10)) )
       	    && (opcode[3:2] != 2'b10) && (data_req_W == 1'b1) )
        	  next_state = decode_instr;
	    //Case 2 instr reads in data but the prev instr is using the fsl
	else if ( ((exec_prog == 1'b0) && (opcode[0] == 1'b1))  
       	    && (opcode[3:2] != 2'b10) && (rx_req_W == 1'b1) )
        	  next_state = decode_instr;
	   //There are no shared resource issues so execute
	else
		  next_state = execute_instr;
      end
    execute_instr:
      begin
	  //Continue executing until the instruction/address has been
	  //completed- that data can be written on its own.
	if ((IR_req_W0 == 1'b1) || (IR_req_W1 == 1'b1) || 
	   (Addr_req_W == 1'b1) || (change_a0 == 1'b1) || 
	   //(two_part_instr == 1'b1) || 
	   (data_req_W == 1'b1) || //added to remove parallelism
	   ((opcode == `wait_rx) && (rx_IR_full == 1'b0)))
		  next_state = execute_instr;
	else
		  next_state = get_next_instr;
        end
  endcase
end

always @ (mux_select_fifo_in or IR_req_W0 or IR_req_W1 or Addr_req_W or data_req_W or fifo_full or valid_data or Hmid)begin
  case(mux_select_fifo_in)
    //HAVE ADDED A (fifo_write == 1'b0) condition so that there are no back
    //to back writes****************
     2'b00:
       begin
         if ((IR_req_W0 == 1'b1) && ((data_req_W == 1'b0) ||
		((data_req_W == 1'b1) && (Hmid == 1'b0)))) 
		//DO I need the last 3 conditions?
		// (fifo_exists == 1'b0)) // && 
		 //(fifo_write == 1'b0))
		 // (pres_state == execute_instr) &&
		// || (IR_req_W1 == 1'b1)) && (ex_IR_full == 1'b1) && 
		 //(fifo_full == 1'b0)) 
            fifo_write = 1'b1;
	 else 
            fifo_write = 1'b0;
       end
     2'b01:
       begin
         if ((IR_req_W1 == 1'b1) && (fifo_full == 1'b0)) // && (fifo_write == 1'b0))
            fifo_write = 1'b1;
	 else 
            fifo_write = 1'b0;
       end
     2'b10:
       begin
         if ((Addr_req_W == 1'b1) && (fifo_full == 1'b0)) // && 
		 //(fifo_write == 1'b0))
            fifo_write = 1'b1;
	 else 
            fifo_write = 1'b0;
       end
     2'b11:
       begin
         if ((data_req_W == 1'b1) && (fifo_full == 1'b0) && 
		 (valid_data == 1'b1)) // && (fifo_write == 1'b0))
            fifo_write = 1'b1;
	 else 
            fifo_write = 1'b0;
       end
  endcase
end

// FSM outputs
//always @ (posedge clk or posedge rst) begin
  //if (rst == 1'b1) begin
    //fifo_write <= 1'b0;
    //pixel_ct <= 10'd0;
  //end
  //else begin
    //case (pres_state) 
      //st_reset:
        //begin
	  //fifo_write <= 1'b0;
	  //pixel_ct <= 10'd0;
        //end
      //st_ctrl:
        //begin
	  //fifo_write <= 1'b1;
	  //pixel_ct <= pixel_ct;
        //end
      //st_ctrl2:
        //begin
	  //fifo_write <= 1'b0;
	  //pixel_ct <= pixel_ct;
        //end
      //st_addr:
        //begin
	  //fifo_write <= 1'b1;
	  //pixel_ct <= pixel_ct;
        //end
      //st_addr2:
        //begin
	  //fifo_write <= 1'b0;
	  //pixel_ct <= pixel_ct;
        //end
      //st_data:
        //begin
	  //fifo_write <= 1'b1;
	  //pixel_ct <= pixel_ct + 1;
       //end
      //st_data2:
        //begin
	  //fifo_write <= 1'b0;
	  //pixel_ct <= pixel_ct;
       //end
    //endcase
  //end
//end

// for debug
// Light an LED if we go back to reset before we've written pixels 0-719
// (or is it 1-720?)
assign led0_o = Fmid; //Vmid; // control_reset; //video_in_on;
assign led1_o = (pres_v_state == st_reset) ? 1'b0 : 1'b1;
/*always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    led0_o <= 1'b1; // off
  //else if (a0_reg[0] == 1'b1)
    //if ((pixel_ct < (`line_length - 1)) && (pres_state != st_reset) && (next_state == st_reset))
      //led0_o <= 1'b0; //on
  else if (a0_reg[1] == 1'b1)
      led0_o <= 1'b0; //off
  else
      //led0_o <= led0_o;
      led0_o <= 1'b1;
end
*/
/*
// Light an LED if we write to a full fifo
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1)
    led1_o <= 1'b1; // off
  //else if (fifo_full == 1'b1)
  else if (batch == 1'b0)
      led1_o <= 1'b0; //on
  //else if (fifo_full == 1'b0)
  else if (batch == 1'b1)
      led1_o <= 1'b1; //off
  else
      led1_o <= led1_o;
end
*/
// Connect internal FIFO to FSL
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) begin
    fifo_read <= 1'b0;
    FSL_M_Writed <= 1'b0;
    wrote_last <= 1'b0;
  end
  else begin
    if ((wrote_last == 1'b0) && (fifo_exists == 1'b1) && (FSL_M_Full == 1'b0)) begin
      // Write to FSL
      fifo_read <= 1'b1;
      FSL_M_Writed <= 1'b1;
      wrote_last <= 1'b1;
    end 
    else begin
      fifo_read <= 1'b0;
      FSL_M_Writed <= 1'b0;
      wrote_last <= 1'b0;
    end
  end
end

// Delay the write b/c the Coregen FIFO has latency at output
always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) 
    FSL_M_Write <= 1'b0;
  else
    FSL_M_Write <= FSL_M_Writed;
end


//always @ (posedge clk or posedge control_reset) begin
  //if (control_reset == 1'b1) 
    //Addr_prev <= 19'b0;
  //else 
    //Addr_prev <= Addr;
//end

always @ (posedge clk or posedge control_reset) begin
  if (control_reset == 1'b1) 
    batch <= 1'b0;
  //else if ((pres_v_state == st_even) && (Vmid == 1'b1))
    ////else if ((Addr == 19'h0) && (Addr_prev != 19'h0))  // frame address resetting
    else if ((Fmid == 1'b0) && (Vmid == 1'b1))
    ////else if (pres_v_state == st_blankA)
    batch <= 1'b0;
  else // if (Vmid == 1'b0)
    batch <= 1'b1;
  //else
    //batch <= batch;
end
	
// Add a FIFO so that data does not get lost if ZBT controller is busy
// servicing the SVGA controller when the video starts up.  This would
// hopefully avoid data loss due to FSL being full, but only for the
// duration of 1 SVGA line (`H_ACTIVE) 
// Has to carry the write & control bit too
/*Sync_FIFO_SIM 
 #(.C_IMPL_STYLE(0),
//   .WordSize(33),
//   .MemSize(`H_ACTIVE))  // impl style (0 => not BRAMs), width, depth
   .WordSize(25),  // Don't need the upper-most word
   .MemSize(730))  // impl style (0 => not BRAMs), width, depth

   sync_fifo (
    .Reset(rst), 
    .Clk(clk), 
    .WE(fifo_write), 
    .DataIn(DataIn), 
    .Full(fifo_full), 
    .RD(fifo_read), 
    .DataOut(DatatoFSL), 
    .Exists(fifo_exists)
    );
*/
coregen_fifo coregen_fifo (
	.clk(clk),
	.sinit(control_reset),
	.din(DataIn),
	.wr_en(fifo_write),
	.rd_en(fifo_read),
	.dout(DatatoFSL),
	.full(fifo_full),
	.empty(fifo_empty)); 


//addr_gen addr_gen (
    //.clk(clk), 
    //.rst(control_reset), 
    //.Fin(Fmid), 
    //.Vin(Vmid), 
    //.Hin(Hmid), 
    //.Addr(Addr)
    //);

//input2rgb input2rgb (
    //.video_bits(VIDEO_INPUT), 
    //.clk(clk), 
    //.rst(rst), 
    //.R(RGBdata[23:16]), 
    //.G(RGBdata[15:8]), 
    //.B(RGBdata[7:0]), 
    //.Fmid(Fmid), 
    //.Vmid(Vmid), 
    //.Hmid(Hmid)
    //);
//


fsl_video_program fsl_video_program (
  // inputs
  .clk(clk), 
  .rst(control_reset), 
  
  //Program Interface:
  .prog_instr(prog_instr),
  .program_cbit(program_cbit),
  .prog_instr_read(prog_instr_read),
  .valid_instr(prog_IR_full),
  .done_field(done_field), //Status_bit 0
  .done_frame(done_frame), //Status_bit 1
  .video_in_on(video_in_on), //Status_bit 2
  .Hmid(Hmid) //Status bit 3
);

video_input video_input(
	.clk(clk),
	.rst(control_reset),
	.VIDEO_INPUT(VIDEO_INPUT),
	//.FSL_M_Full(FSL_M_Full),

	//added to "new" video_input
	//.fifo_read(fifo_read),
	//.fifo_empty(fifo_empty),
	//.DatatoFSL(DatatoFSL),

//	.release_priority(release_priority),
//	.FSL_M_Clk(FSL_M_Clk),
	//.FSL_M_Data(FSL_M_Data),
	//.FSL_M_Control(FSL_M_Control),
	//.FSL_M_Write(FSL_M_Write),
	//.led0_o(led0_o),
	//.led1_o(led1_o),
	//.batch(batch)
        .R_data(RGBdata[23:16]), 
        .G_data(RGBdata[15:8]), 
        //.B_data(RGBdata[7:0]) 
	//REMOVED FOR TESTING ONLY
        .B_data(RGBdata[7:0]), 
	.Fmid(Fmid),
	.Vmid(Vmid),
	.Hmid(Hmid)
	);

endmodule



