//************************************************************************
//    This file was written for University of Toronto, ECE532 
//    File: fsl_video_program.v
//    Used by: fsl_video_input.v
//    Works with ProjectNavigator 6.2.03i
//    Description: Provides the instructions to be executed by the video
//    controller.  It should only be necessary for a user to change the 
//    program to suit the application- the controller should remain
//    unchanged
//
//
//    LLS October 2004
//************************************************************************ 
//
 

module fsl_video_program (
  // inputs
  clk, rst, 
  
  //Program Interface:
  prog_instr,
  program_cbit,
  prog_instr_read,
  valid_instr,  //only when high should the instruction be read (use to stall)
  done_field, //Status_bit 0
  done_frame, //Status_bit 1
  video_in_on, //Status_bit 2
  Hmid  //Status bit 3- indicates new line coming- 4 clock cycles early to 
  		//allow for processing
  );

  input clk, rst;
  output [0:31] prog_instr;  //Next instruction to be executed by the controller

  output program_cbit;
  output valid_instr;
  input prog_instr_read; //Read the instruction on the prog_instr bus
  input done_field; //Status_bit 0
  input done_frame; //Status_bit 1
  input video_in_on; //Status_bit 2
  input Hmid; //Status bit 3
	  
  wire clk, rst;
 
  //Program interface:
  wire [0:31] prog_instr;
  wire program_cbit;
  wire prog_instr_read; 
  wire done_field;
  wire done_frame;
  wire Hmid;
  reg Hmid_prev;
  reg valid_instr; 

  reg done_field_prev;
  reg done_frame_prev;
  reg current_field;
  reg current_frame;
  reg change_a0;
  reg [2:0] pres_state, next_state;
  reg [0:31] int_prog_instr;
  reg int_prog_cbit;

////Test instructions:
//`define write_imm 32'h00001003
//`define write_abs_imm 32'h00000f07
//`define abs_addr 32'h00440033
//`define write_a0 32'h000000B
//`define a0_new_addr 32'h001100AA
//`define add_a0 32'h00220018
//`define sub_a0 32'h00007708
//`define write_indirect 32'h0000140F
//`define write_autoinc 32'h0000111F
//`define noop 32'h00000004
//`define reset 32'h00000000
//`define wait_rx 32'h0000000C

////TEST STATES
//// NOTE: The states reflect the program structure
//parameter  st_wr_imm=0;
//parameter  st_wr_abs_imm=1;
//parameter  st_wr_addr=2;
//parameter  st_wr_a0=3;
//parameter  st_wr_new_a0=4;
//parameter  st_add_a0=5;
//parameter  st_sub_a0=6;
//parameter  st_wr_indirect=7;
//parameter  st_wr_autoinc=8;
//parameter  st_noop=9;
//parameter  st_reset=10;
//parameter  st_wait_rx=11;

// NOTE: The states reflect the program structure
parameter  st_wr_addr_a0=0;
parameter  st_wr_odd_addr_bank0=1;
parameter  st_wr_even_addr_bank0=2;
parameter  st_wr_odd_addr_bank1=3;
parameter  st_wr_even_addr_bank1=4;
parameter  st_wr_data_a0_inc=5;
parameter  st_add_line_to_a0=6;
////parameter  st_wr_line=7;  //UNECESSARY

////`define control_word 32'h00028007 //Used by the SIMPPL protocol
////`define control_word 32'h00001007 //Used by the SIMPPL protocol
////`define line_length 10'd640 
////ACTUAL PROGRAM:
`define wr_addr_a0 32'h0000000B //Write base address to a0
`define addr_bank0_odd 32'h00000000 //Odd Base address for bank 0
`define addr_bank0_even 32'h00000280 //Even Base address for bank 0
`define addr_bank1_odd 32'h80000000 //Odd address for bank 1
`define addr_bank1_even 32'h80000280 //Even address for bank 1
`define wr_line_a0_inc 32'h0002801F //Write 640 words to a0 and post increment
`define add_a0_line 32'h00028018 //add another line to the count to skip
					//to next line
//`define addr_bank0_even 32'h00000140 //Even Base address for bank 0
//`define addr_bank1_odd 32'h80000000 //Odd address for bank 1
//`define addr_bank1_even 32'h80000140 //Even address for bank 1
//`define wr_line_a0_inc 32'h0001401F //Write 640 words to a0 and post increment
//`define add_a0_line 32'h00014018 //add another line to the count to skip
//**FOR TESTING PROGRAM:
//`define wr_addr_a0 32'h0000000B //Write base address to a0
//`define addr_bank0_odd 32'h00000000 //Odd Base address for bank 0
//`define addr_bank0_even 32'h00000004 //Even Base address for bank 0
//`define addr_bank1_odd 32'h80000000 //Odd address for bank 1
//`define addr_bank1_even 32'h80000004 //Even address for bank 1
//`define wr_line_a0_inc 32'h0000041F //Write 640 words to a0 and post increment
//`define add_a0_line 32'h00000418 //add another line to the count to skip
					//to next line

assign prog_instr = int_prog_instr;
assign program_cbit = int_prog_cbit;

// Hmid previous value:
always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
    Hmid_prev <= 1'b0;
  else
    Hmid_prev <= Hmid;
end

// done_frame previous value:
always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
    done_frame_prev <= 1'b0;
  else
    done_frame_prev <= done_frame;
end

// done_field previous value:
always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
    done_field_prev <= 1'b0;
  else
    done_field_prev <= done_field;
end

always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
    change_a0 <= 1'b0;
  else if ((video_in_on == 1'b1) && 
	  (done_field == 1'b1) && (done_field_prev == 1'b0))
    change_a0 <= 1'b1;
  else if (pres_state == st_wr_addr_a0)
  //else if (done_field == 1'b0)
    change_a0 <= 1'b0;
  else 
    change_a0 <= change_a0;
end

// Stores the current field:
always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
    current_field <= 1'b0;
  else if ((video_in_on == 1'b1) && 
	  (done_field == 1'b1) && (done_field_prev == 1'b0))
    current_field <= ~current_field;
  else
    current_field <= current_field;
end

// Stores the current frame:
always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
    current_frame <= 1'b0;
  else if ((video_in_on == 1'b1) && 
	  (done_frame == 1'b1) && (done_frame_prev == 1'b0))
    current_frame <= ~current_frame;
  else
    current_frame <= current_frame;
end

//Used to indicate when the instruction is valid.  Allows us to stall the
//processor when we don't know for sure the next state
always @ (pres_state or Hmid or Hmid_prev) begin
  case(pres_state)
    st_wr_addr_a0:
      valid_instr = 1'b1;
    st_wr_odd_addr_bank0:
      valid_instr = 1'b1;
    st_wr_even_addr_bank0:
      valid_instr = 1'b1;
    st_wr_odd_addr_bank1:
      valid_instr = 1'b1;
    st_wr_even_addr_bank1:
      valid_instr = 1'b1;
    st_wr_data_a0_inc:
      begin
	if ((Hmid == 1'b0) && (Hmid_prev == 1'b1))
          valid_instr = 1'b1;
	else 
          valid_instr = 1'b0;
      end
    st_add_line_to_a0:
      valid_instr = 1'b1;
  endcase
end

// FSM register
always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1)
    pres_state <= st_wr_addr_a0;
    //pres_state <= st_wr_imm;
  else
    pres_state <= next_state;
end

//Next-state state machine for the program:
always @ (pres_state or prog_instr_read or change_a0 or current_frame or current_field or rst) 
begin

  case (pres_state)
    st_wr_addr_a0:
      begin
      	if ((prog_instr_read == 1'b1) && (rst == 1'b0)) begin
	  if ((current_frame == 1'b0) && (current_field == 1'b0))
            next_state = st_wr_odd_addr_bank0;
	  else if ((current_frame == 1'b0) && (current_field == 1'b1))
            next_state = st_wr_even_addr_bank0;
	  else if ((current_frame == 1'b1) && (current_field == 1'b0))
            next_state = st_wr_odd_addr_bank1;
	  else if ((current_frame == 1'b1) && (current_field == 1'b1))
            next_state = st_wr_even_addr_bank1;
          end
        else
	  next_state = st_wr_addr_a0;
      end
    st_wr_odd_addr_bank0:
      begin 
	if (prog_instr_read == 1'b1) 
          next_state = st_wr_data_a0_inc;
	else
          next_state = st_wr_odd_addr_bank0;
      end
    st_wr_even_addr_bank0:
      begin 
	if (prog_instr_read == 1'b1) 
          next_state = st_wr_data_a0_inc;
	else
          next_state = st_wr_even_addr_bank0;
      end
    st_wr_odd_addr_bank1:
      begin 
	if (prog_instr_read == 1'b1)
          next_state = st_wr_data_a0_inc;
	else
          next_state = st_wr_odd_addr_bank1;
      end
    st_wr_even_addr_bank1:
      begin 
	if (prog_instr_read == 1'b1)
          next_state = st_wr_data_a0_inc;
	else
          next_state = st_wr_even_addr_bank1;
      end
    st_wr_data_a0_inc:
      begin 
        if (change_a0 == 1'b1)
	  next_state = st_wr_addr_a0;
	else if (prog_instr_read == 1'b1) 
	  next_state = st_add_line_to_a0;
        else
          next_state = st_wr_data_a0_inc;
      end
    st_add_line_to_a0:
      begin 
        if (change_a0 == 1'b1)
	  next_state = st_wr_addr_a0;
	else if (prog_instr_read == 1'b1) 
	  next_state = st_wr_data_a0_inc;
        else
          next_state = st_add_line_to_a0;
      end
    default:
      next_state = st_wr_addr_a0;
  endcase
end

//Sets the current instruction output to the controller based on the state:
always @ (posedge clk or posedge rst) begin
  if (rst == 1'b1) 
    begin
      int_prog_instr <= `wr_addr_a0;
      int_prog_cbit <= 1'b1;
    end	  
  else
    begin
      case (next_state) 
        st_wr_addr_a0:
          begin
      	    int_prog_instr <= `wr_addr_a0;
	    int_prog_cbit <= 1'b1;
          end
        st_wr_odd_addr_bank0:
          begin
      	    int_prog_instr <= `addr_bank0_odd;
	    int_prog_cbit <= 1'b0;
          end
        st_wr_even_addr_bank0:
          begin
      	    int_prog_instr <= `addr_bank0_even;
	    int_prog_cbit <= 1'b0;
          end
        st_wr_odd_addr_bank1:
          begin
      	    int_prog_instr <= `addr_bank1_odd;
	    int_prog_cbit <= 1'b0;
          end
        st_wr_even_addr_bank1:
          begin
      	    int_prog_instr <= `addr_bank1_even;
	    int_prog_cbit <= 1'b0;
          end
        st_wr_data_a0_inc:
          begin
      	    int_prog_instr <= `wr_line_a0_inc;
	    int_prog_cbit <= 1'b1;
          end
        st_add_line_to_a0:
          begin
      	    int_prog_instr <= `add_a0_line;
	    int_prog_cbit <= 1'b1;
          end
        default:
          begin
      	    int_prog_instr <= `wr_addr_a0;
	    int_prog_cbit <= 1'b1;
          end
      endcase
    end
end

////Next state state machine for test program:
//always @ (pres_state or prog_instr_read or done_frame or rst) begin
  //case (pres_state) 
    //st_wr_imm:
      //begin
        //if ((prog_instr_read == 1'b1) && (rst == 1'b0))
		//next_state <= st_wr_abs_imm;
        //else
		//next_state <= st_wr_imm;
      //end
    //st_wr_abs_imm:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_wr_addr;
        //else
		//next_state <= st_wr_abs_imm;
      //end
    //st_wr_addr:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_wr_a0;
        //else
		//next_state <= st_wr_addr;
      //end
    //st_wr_a0:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_wr_new_a0;
        //else
		//next_state <= st_wr_a0;
      //end
    //st_wr_new_a0:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_add_a0;
        //else
		//next_state <= st_wr_new_a0;
      //end
    //st_add_a0:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_sub_a0;
        //else
		//next_state <= st_add_a0;
      //end
    //st_sub_a0:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_wr_indirect;
        //else
		//next_state <= st_sub_a0;
      //end
    //st_wr_indirect:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_wr_autoinc;
        //else
		//next_state <= st_wr_indirect;
      //end
    //st_wr_autoinc:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_noop;
        //else
		//next_state <= st_wr_autoinc;
      //end
    //st_noop:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_reset;
        //else
		//next_state <= st_noop;
      //end
    //st_reset:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_wait_rx;
        //else
		//next_state <= st_reset;
      //end
    //st_wait_rx:
      //begin
        //if (prog_instr_read == 1'b1)
		//next_state <= st_wr_imm;
        //else
		//next_state <= st_wait_rx;
      //end
    //default:
      //next_state <= st_wr_imm;
  //endcase
//end

////TEST PROGRAM:
////Sets the current instruction output to the controller based on the state:
//always @ (posedge clk or posedge rst) begin
  //case (next_state) 
    //st_wr_imm:
      //begin
      	//int_prog_instr = `write_imm;
	//int_prog_cbit = 1'b1;
      //end
    //st_wr_abs_imm:
      //begin
      	//int_prog_instr = `write_abs_imm;
	//int_prog_cbit = 1'b1;
      //end
    //st_wr_addr:
      //begin
      	//int_prog_instr = `abs_addr;
	//int_prog_cbit = 1'b0;
      //end
    //st_wr_a0:
      //begin
      	//int_prog_instr = `write_a0;
	//int_prog_cbit = 1'b1;
      //end
    //st_wr_new_a0:
      //begin
      	//int_prog_instr = `a0_new_addr;
	//int_prog_cbit = 1'b0;
      //end
    //st_add_a0:
      //begin
      	//int_prog_instr = `add_a0;
	//int_prog_cbit = 1'b1;
      //end
    //st_sub_a0:
      //begin
      	//int_prog_instr = `sub_a0;
	//int_prog_cbit = 1'b1;
      //end
    //st_wr_indirect:
      //begin
      	//int_prog_instr = `write_indirect;
	//int_prog_cbit = 1'b1;
      //end
    //st_wr_autoinc:
      //begin
      	//int_prog_instr = `write_autoinc;
	//int_prog_cbit = 1'b1;
      //end
    //st_noop:
      //begin
      	//int_prog_instr = `noop;
	//int_prog_cbit = 1'b1;
      //end
    //st_reset:
      //begin
      	//int_prog_instr = `reset;
	//int_prog_cbit = 1'b1;
      //end
    //st_wait_rx:
      //begin
      	//int_prog_instr = `wait_rx;
	//int_prog_cbit = 1'b1;
      //end
    //default:
      //begin
      	//int_prog_instr = `write_imm;
	//int_prog_cbit = 1'b1;
      //end
  //endcase
//end

endmodule
