--************************************************************************
--    This file was written for University of Toronto, ECE532 
--    File: fsl_zbt_control.v
--    Uses: fsl_zbt_control_core.v
--    Used by: 
--    For ProjectNavigator 6.2.03i
--    Description: Module to wrap zbt controller for use with 2 pairs of 
--    FSL connections.  This version of the ZBT controller is for the case 
--    when synchronization is *NOT* critical for both devices attached to the
--    ZBT controller
--
--    Might still need to be debugged a bit
--
--    NCKC August 2004
--************************************************************************ 
--
-- FSL_Clk should be twice as fast as Logic_Clk, so use DCM outside
--
-- Also need to make sure Logic_Clk is running at same frequency as 
-- user_memory_ram_clock
-- RAM clock should be generated by DCM with external feedback pins on the 
-- FPGA connected.
--
-- See CLOCKGEN example provided by Xilinx


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.STD_Logic_arith.all;
use IEEE.STD_Logic_unsigned.all;

entity FSL_ZBT_CONTROL is
  generic (
    C_DWIDTH      : integer := 32;
    ADDR_BITS     : integer := 19;
    DATA_BITS     : integer := 32;
    ZBT_READ_LATENCY  : integer := 4
  );
  port (

    Can_switch : out std_logic;

    -- Global signals
    FSL_Clk   : in std_logic;	
    Logic_Clk : in std_logic;
    Reset : in std_logic;
    release_priority : in std_logic;

    -- Slave FSL0 Signals
    FSL0_S_Clk     : out std_logic;  
    FSL0_S_Read    : out std_logic;
    FSL0_S_Data    : in  std_logic_vector(0 to C_DWIDTH-1);
    FSL0_S_Control : in  std_logic;
    FSL0_S_Exists  : in  std_logic;

    -- Master FSL0 Signals
    FSL0_M_Clk     : out std_logic;  
    FSL0_M_Write   : out std_logic;
    FSL0_M_Data    : out std_logic_vector(0 to C_DWIDTH-1);
    FSL0_M_Control : out std_logic;
    FSL0_M_Full    : in  std_logic;

    -- Slave FSL1 Signals
    FSL1_S_Clk     : out std_logic;  
    FSL1_S_Read    : out std_logic;
    FSL1_S_Data    : in  std_logic_vector(0 to C_DWIDTH-1);
    FSL1_S_Control : in  std_logic;
    FSL1_S_Exists  : in  std_logic;

    -- Master FSL1 Signals
    FSL1_M_Clk     : out std_logic;  
    FSL1_M_Write   : out std_logic;
    FSL1_M_Data    : out std_logic_vector(0 to C_DWIDTH-1);
    FSL1_M_Control : out std_logic;
    FSL1_M_Full    : in  std_logic;

    -- ZBT signals
    MEMORY0_CLK_P   : out std_logic;
    MEMORY0_CLKEN_N : out std_logic;
    MEMORY0_WEN_N   : out std_logic;
    MEMORY0_WENA_N  : out std_logic;
    MEMORY0_WENB_N  : out std_logic;
    MEMORY0_WENC_N  : out std_logic;
    MEMORY0_WEND_N  : out std_logic;
    MEMORY0_ADV_LD_N : out std_logic;
    MEMORY0_OEN_N   : out std_logic; 
    MEMORY0_CEN_N   : out std_logic;
    MEMORY0_ADDR_P  : out std_logic_vector(ADDR_BITS-1 downto 0);
    MEMORY0_DATA_A_P : inout std_logic_vector(7 downto 0);
    MEMORY0_DATA_B_P : inout std_logic_vector(7 downto 0);
    MEMORY0_DATA_C_P : inout std_logic_vector(7 downto 0);
    MEMORY0_DATA_D_P : inout std_logic_vector(7 downto 0);
    
    user_memory_ram_clock_zbt : in std_logic;
    user_memory_clocken_zbt   : in std_logic;  -- I assume we'd just want this to be '1'
    user_memory_load_addr_zbt : in std_logic;  -- I assume we'd just want this to be '1'

    user_memory_ce_zbt        : in std_logic  -- I assume we'd just want this to be '1'


    );


end entity FSL_ZBT_CONTROL;

library unisim;
use unisim.all;


architecture IMP of FSL_ZBT_CONTROL is
  component FSL_ZBT_CONTROL_CORE is
  generic (
    C_DWIDTH      : integer := 32;
    ADDR_BITS     : integer := 19;
    DATA_BITS     : integer := 32;
    ZBT_READ_LATENCY  : integer := 4
  );
  port (
    FSL_Clk : in std_logic;
    Logic_Clk : in std_logic;
    Reset : in std_logic;
    Can_Switch_FSLs_o : out std_logic;

    -- Slave FSL Signals
    FSL_S_Read_o    : out std_logic;
    FSL_S_Data_i    : in  std_logic_vector(0 to C_DWIDTH-1);
    FSL_S_Control_i : in  std_logic;
    FSL_S_Exists_i  : in  std_logic;

    -- Master FSL Signals
    FSL_M_Write_o   : out std_logic;
    FSL_M_Data_o    : out std_logic_vector(0 to C_DWIDTH-1);
    FSL_M_Control_o : out std_logic;
    FSL_M_Full_i    : in  std_logic;

    -- ZBT signals
    MEMORY0_CLK_P_o   : out std_logic;
    MEMORY0_CLKEN_N_o : out std_logic;
    MEMORY0_WEN_N_o   : out std_logic;
    MEMORY0_WENA_N_o  : out std_logic;
    MEMORY0_WENB_N_o  : out std_logic;
    MEMORY0_WENC_N_o  : out std_logic;
    MEMORY0_WEND_N_o  : out std_logic;
    MEMORY0_ADV_LD_N_o : out std_logic;
    MEMORY0_OEN_N_o   : out std_logic; 
    MEMORY0_CEN_N_o   : out std_logic;
    MEMORY0_ADDR_P_o  : out std_logic_vector(ADDR_BITS-1 downto 0);
    MEMORY0_DATA_A_P_io : inout std_logic_vector(7 downto 0);
    MEMORY0_DATA_B_P_io : inout std_logic_vector(7 downto 0);
    MEMORY0_DATA_C_P_io : inout std_logic_vector(7 downto 0);
    MEMORY0_DATA_D_P_io : inout std_logic_vector(7 downto 0);
    
    user_memory_ram_clock_zbt_i : in std_logic;
    user_memory_clocken_zbt_i   : in std_logic;  -- I assume we'd just want this to be '1'
    user_memory_load_addr_zbt_i : in std_logic;  -- I assume we'd just want this to be '1'
    user_memory_ce_zbt_i        : in std_logic  -- I assume we'd just want this to be '1'
  );
  end component FSL_ZBT_CONTROL_CORE;
  
  signal FSL_in_select : std_logic;
  signal FSL_S_Read : std_logic;
  signal FSL_S_Data : std_logic_vector(0 to C_DWIDTH-1);
  signal FSL_S_Control : std_logic;
  signal FSL_S_Exists: std_logic;

  signal FSL_out_select : std_logic;
  signal FSL_M_Write : std_logic;
  signal FSL_M_Data : std_logic_vector(0 to C_DWIDTH-1);
  signal FSL_M_Control : std_logic;
  signal FSL_M_Full: std_logic;

  signal Can_Switch_FSLs : std_logic;

  signal Internal_reset : std_logic;
  signal priority_reset : std_logic;
  signal release_priorityregs : std_logic_vector(1 downto 0);

--  signal delsw : std_logic_vector(3 downto 0);
  signal delsw : std_logic_vector(1 downto 0);
  signal delok : std_logic;

  signal ctrl_exist : std_logic; -- For LED debugging


begin  -- architecture IMP

--  Can_switch <= Can_Switch_FSLs; 
  ctrl_exist <= FSL_S_Control and FSL_S_Exists;

  Internal_reset <= (Reset or priority_reset);

  FSL0_S_Clk <= FSL_Clk;
  FSL0_M_Clk <= FSL_Clk;
  FSL1_S_Clk <= FSL_Clk;
  FSL1_M_Clk <= FSL_Clk;


  -- Deal with signals from multiple FSL's
  with FSL_in_select select
    FSL0_S_Read <= FSL_S_Read when '0',
                 '0' when others;

  with FSL_in_select select
    FSL1_S_Read <= FSL_S_Read when '1',
                 '0' when others;

  with FSL_in_select select
    FSL_S_Data <= FSL0_S_Data when '0',
                  FSL1_S_Data when others;

  with FSL_in_select select
    FSL_S_Control <= FSL0_S_Control when '0',
                     FSL1_S_Control when others;

  with FSL_in_select select
    FSL_S_Exists <= FSL0_S_Exists when '0',
                     FSL1_S_Exists when others;

  with FSL_out_select select
    FSL0_M_Write <= FSL_M_Write when '0',
                 '0' when others;

  with FSL_out_select select
    FSL1_M_Write <= FSL_M_Write when '1',
                 '0' when others;

  FSL0_M_Data <= FSL_M_Data when FSL_out_select = '0' else (others => '0');

  FSL1_M_Data <= FSL_M_Data when FSL_out_select = '1' else (others => '0');

  with FSL_out_select select
    FSL0_M_Control <= FSL_M_Control when '0',
                 '0' when others;

  with FSL_out_select select
    FSL1_M_Control <= FSL_M_Control when '1',
                 '0' when others;

  with FSL_out_select select
    FSL_M_Full <= FSL0_M_Full when '0',
                     FSL1_M_Full when others;

  -- Reset the controller so that we can start servicing the priority side ASAP
  priority_reset <= (not release_priorityregs(0)) and release_priorityregs(1);

  chng_priority: process (Logic_Clk, Reset)
  begin
    if (Reset = '1') then
      release_priorityregs <= "00";
    else
      release_priorityregs(1) <= release_priorityregs(0);
      release_priorityregs(0) <= release_priority;
    end if;
  end process chng_priority;


  -- Selection logic to choose between FSL's for input & output
  -- For more FSLs use a counter
  -- For 2 FSLs it didn't want to synthesize as a 1-bit counter
--  delok <= not (delsw(0) or delsw(1) or delsw(2) or delsw(3));
  delok <= not (delsw(0) or delsw(1));

  fslsel: process (Logic_Clk, Reset)
  begin
    if (Reset = '1') then
      FSL_in_select <= '0';
      FSL_out_select <= '0';
--      delsw <= "0000";
      Can_switch <= '0';
      delsw <= "00";
    elsif (Logic_Clk'event) and (Logic_Clk = '1') then
      if (release_priority = '0') then -- Priority device wants command back
        FSL_in_select <= '0';
        FSL_out_select <= '0';
        Can_switch <= '0';
--        delsw(3 downto 1) <= delsw(2 downto 0);
        delsw(1) <= delsw(0);
        delsw(0) <= '1';
      elsif (Can_Switch_FSLs = '1') and (delok = '1') then  -- Didn't switch in the last 2 cycles already
        if (ctrl_exist = '1' ) then
          Can_switch <= '1';    -- Light LED to alert that Can_Switch_FSLs was still high when ctrl existed -> bad
        else
          Can_switch <= '0';
        end if;
        FSL_in_select <= not FSL_in_select;
        FSL_out_select <= not FSL_out_select;
--        delsw(3 downto 1) <= delsw(2 downto 0);
        delsw(1) <= delsw(0);
        delsw(0) <= '1';
      else
        FSL_in_select <= FSL_in_select;
        FSL_out_select <= FSL_out_select;
--        delsw(3 downto 1) <= delsw(2 downto 0);
        delsw(1) <= delsw(0);
        delsw(0) <= '0';
        Can_switch <= '0';
      end if;
    end if;   
  end process fslsel;

  -- Instantiate the core of the FSL zbt controller
  FSL_ZBT_Inst : FSL_ZBT_CONTROL_CORE 
    generic map (
    C_DWIDTH      => 32,
    ADDR_BITS     => 19,
    DATA_BITS     => 32,
    ZBT_READ_LATENCY  => 4
    )
  port map (
    FSL_Clk => FSL_Clk,
    Logic_Clk => Logic_Clk,
    Reset => Internal_reset,
    Can_Switch_FSLs_o => Can_Switch_FSLs,

    -- Slave FSL Signals
    FSL_S_Read_o    => FSL_S_Read,
    FSL_S_Data_i    => FSL_S_Data,
    FSL_S_Control_i => FSL_S_Control,
    FSL_S_Exists_i  => FSL_S_Exists,

    -- Master FSL Signals
    FSL_M_Write_o   => FSL_M_Write,
    FSL_M_Data_o    => FSL_M_Data,
    FSL_M_Control_o => FSL_M_Control,
    FSL_M_Full_i    => FSL_M_Full,

    -- ZBT signals
    MEMORY0_CLK_P_o   => MEMORY0_CLK_P,
    MEMORY0_CLKEN_N_o => MEMORY0_CLKEN_N,
    MEMORY0_WEN_N_o   => MEMORY0_WEN_N,
    MEMORY0_WENA_N_o  => MEMORY0_WENA_N,
    MEMORY0_WENB_N_o  => MEMORY0_WENB_N,
    MEMORY0_WENC_N_o  => MEMORY0_WENC_N,
    MEMORY0_WEND_N_o  => MEMORY0_WEND_N,
    MEMORY0_ADV_LD_N_o => MEMORY0_ADV_LD_N,
    MEMORY0_OEN_N_o   => MEMORY0_OEN_N, 
    MEMORY0_CEN_N_o   => MEMORY0_CEN_N,
    MEMORY0_ADDR_P_o  => MEMORY0_ADDR_P,
    MEMORY0_DATA_A_P_io => MEMORY0_DATA_A_P,
    MEMORY0_DATA_B_P_io => MEMORY0_DATA_B_P,
    MEMORY0_DATA_C_P_io => MEMORY0_DATA_C_P,
    MEMORY0_DATA_D_P_io => MEMORY0_DATA_D_P,
    
    user_memory_ram_clock_zbt_i => user_memory_ram_clock_zbt,
    user_memory_clocken_zbt_i   => user_memory_clocken_zbt,
    user_memory_load_addr_zbt_i => user_memory_load_addr_zbt,
    user_memory_ce_zbt_i        => user_memory_ce_zbt
  );

end architecture IMP;
